# Standard Library Imports
import logging

from collections import defaultdict
from enum import Enum
from typing import Any, Dict

# Protean
import inflection

from protean.utils import DomainObjects, fully_qualified_name

logger = logging.getLogger("protean.domain")


class _DomainRegistry:
    def __init__(self):
        self._elements: Dict[str, dict] = {}

        # Initialize placeholders for element types
        for element_type in DomainObjects:
            self._elements[element_type.value] = defaultdict(dict)

    class DomainRecord:
        def __init__(self, name: str, qualname: str, class_type: str, cls: Any):
            self.name = name
            self.qualname = qualname
            self.class_type = class_type
            self.cls = cls

    def _is_invalid_element_cls(self, element_cls):
        """Ensure that we are dealing with an element class, that:

        * Has a `element_type` attribute
        * `element_type` is an Enum value
        * The value of `element_type` enum is among recognized `DomainObjects` values
        """
        return (
            not hasattr(element_cls, "element_type")
            or not isinstance(element_cls.element_type, Enum)
            or element_cls.element_type.name not in DomainObjects.__members__
        )

    def register_element(self, element_cls):
        if self._is_invalid_element_cls(element_cls):
            raise NotImplementedError

        element_name = fully_qualified_name(element_cls)

        element = self._elements[element_cls.element_type.value][element_name]
        if element:
            # raise ConfigurationError(f'Element {element_name} has already been registered')
            logger.debug(f"Element {element_name} was already in the registry")
        else:
            element_record = _DomainRegistry.DomainRecord(
                name=element_cls.__name__,
                qualname=element_name,
                class_type=element_cls.element_type.value,
                cls=element_cls,
            )

            self._elements[element_cls.element_type.value][
                element_name
            ] = element_record

            logger.debug(
                f"Registered Element {element_name} with Domain as a {element_cls.element_type.value}"
            )

    def delist_element(self, element_cls):
        if self._is_invalid_element_cls(element_cls):
            raise NotImplementedError

        element_name = fully_qualified_name(element_cls)

        self._elements[element_cls.element_type.value].pop(element_name, None)

    def get(self, element_type):
        return self._elements[element_type.value]


# Set up access to all elements as properties
for element_type in DomainObjects:
    """Set up `properties` on Registry

    Since all elements are stored within a Dict in the registry, accessing
    them will mean knowing the storage structure. It is instead preferable to
    expose the elements by their element types as properties.

    Registry object will contain properties named after of each element type
    and pluralized to indicate that all elements of the type will be returned.

    E.g.
    AGGREGATE: registry.aggregates
    VALUE_OBJECT: registry.value_objects
    """

    # Lowercase element type, add underscores and pluralize
    prop_name = inflection.pluralize(inflection.underscore(element_type.value.lower()))

    # This weird syntax is because when using lambdas in a for loop, we need to supply
    #   element_type as an argument with a default value of element_type
    prop = property(
        lambda self, element_type=element_type: self.get(element_type)
    )  # pragma: no cover  # FIXME Is it possible to cover this line in tests

    # Set the property on the class
    setattr(_DomainRegistry, prop_name, prop)



"""Serializer Object Functionality and Classes"""
# Standard Library Imports
import logging

# Protean
from marshmallow import Schema, fields
from protean.core.exceptions import NotSupportedError
from protean.core.field.basic import (
    Boolean,
    Date,
    DateTime,
    Dict,
    Field,
    Float,
    Identifier,
    Integer,
    List,
    Method,
    Nested,
    String,
    Text,
)
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger("protean.application.serializer")


class _SerializerMetaclass(type):
    """
    This base metaclass processes the class declaration and constructs a Marshmellow Class meta object that can
    be used to load and dump data. It also sets up a `meta_` attribute on the Serializer to be an instance of Meta,
    either the default or one that is defined in the Serializer class.

    `meta_` is setup with these attributes:
        * `declared_fields`: A dictionary that gives a list of any instances of `Field`
            included as attributes on either the class or on any of its superclasses
    """

    def __new__(mcs, name, bases, attrs, **kwargs):  # noqa: C901
        """Initialize Serializer MetaClass and load attributes"""

        def _declared_base_class_fields(bases, attrs):
            """If this class is subclassing another Serializer, add that Serializer's
            fields.  Note that we loop over the bases in *reverse*.
            This is necessary in order to maintain the correct order of fields.
            """
            declared_fields = {}

            for base in reversed(bases):
                if hasattr(base, "meta_") and hasattr(base.meta_, "declared_fields"):
                    base_class_fields = {
                        field_name: field_obj
                        for (
                            field_name,
                            field_obj,
                        ) in base.meta_.declared_fields.items()
                        if field_name not in attrs and not field_obj.identifier
                    }
                    declared_fields.update(base_class_fields)

            return declared_fields

        def _declared_fields(attrs):
            """Load field items into Class"""
            declared_fields = {}

            for attr_name, attr_obj in attrs.items():
                if isinstance(attr_obj, Field):
                    declared_fields[attr_name] = attr_obj

            return declared_fields

        # Ensure initialization is only performed for subclasses of Serializer
        # (excluding Serializer class itself).
        parents = [b for b in bases if isinstance(b, _SerializerMetaclass)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Load declared fields
        declared_fields = _declared_fields(attrs)

        # Load declared fields from Base class, in case this Entity is subclassing another
        base_class_fields = _declared_base_class_fields(bases, attrs)

        all_fields = {**declared_fields, **base_class_fields}

        schema_fields = {}
        for field_name, field_obj in all_fields.items():
            if isinstance(field_obj, Boolean):
                schema_fields[field_name] = fields.Boolean()
            elif isinstance(field_obj, Date):
                schema_fields[field_name] = fields.Date()
            elif isinstance(field_obj, DateTime):
                schema_fields[field_name] = fields.DateTime()
            elif isinstance(field_obj, Identifier):
                schema_fields[field_name] = fields.String()
            elif isinstance(field_obj, String):
                schema_fields[field_name] = fields.String()
            elif isinstance(field_obj, Text):
                schema_fields[field_name] = fields.String()
            elif isinstance(field_obj, Integer):
                schema_fields[field_name] = fields.Integer()
            elif isinstance(field_obj, Float):
                schema_fields[field_name] = fields.Float()
            elif isinstance(field_obj, Method):
                schema_fields[field_name] = fields.Method(field_obj.method_name)
            elif isinstance(field_obj, List):
                schema_fields[field_name] = fields.List(
                    fields.String()
                )  # FIXME Accept type param in List field
            elif isinstance(field_obj, Dict):  # FIXME Accept type param in Dict field
                schema_fields[field_name] = fields.Dict(keys=fields.Str())
            elif isinstance(field_obj, Nested):
                schema_fields[field_name] = fields.Nested(
                    field_obj.schema_name, many=field_obj.many
                )
            else:
                raise NotSupportedError(
                    "{} Field not supported".format(type(field_obj))
                )

        # Remove Protean fields from Serializer class
        for field_name in schema_fields:
            attrs.pop(field_name, None)

        # Update `attrs` with new marshmallow fields
        attrs.update(schema_fields)

        # Remove `abstract` in base classes if defined
        for base in bases:
            if hasattr(base, "Meta") and hasattr(base.Meta, "abstract"):
                delattr(base.Meta, "abstract")

        # Explicit redefinition element_type  necessary because `attrs`
        #   are reset when a serializer class is initialized.
        attrs["element_type"] = DomainObjects.SERIALIZER

        new_class = type(name, (Schema,), attrs)

        # Gather `Meta` class/object if defined
        attr_meta = attrs.pop("Meta", None)
        meta = attr_meta or getattr(new_class, "Meta", None)
        setattr(new_class, "meta_", SerializerMeta(meta, declared_fields))

        return new_class


class SerializerMeta:
    """ Metadata info for the Serializer.

    Also acts as a placeholder for generated entity fields like:

        :declared_fields: dict
            Any instances of `Field` included as attributes on either the class
            or on any of its superclasses will be include in this dictionary.
    """

    def __init__(self, meta, declared_fields):
        self.aggregate_cls = getattr(meta, "aggregate_cls", None)

        # Initialize Options
        self.declared_fields = declared_fields if declared_fields else {}


class BaseSerializer(metaclass=_SerializerMetaclass):
    """The Base class for Protean-Compliant Serializers.

    Provides helper methods to load and dump data during runtime, from protean entity objects.

    Basic Usage::

        @Serializer
        class Address:
            unit = field.String()
            address = field.String(required=True, max_length=255)
            city = field.String(max_length=50)
            province = field.String(max_length=2)
            pincode = field.String(max_length=6)

    (or)

        class Address(BaseSerializer):
            unit = field.String()
            address = field.String(required=True, max_length=255)
            city = field.String(max_length=50)
            province = field.String(max_length=2)
            pincode = field.String(max_length=6)

        domain.register_element(Address)
    """

    def __new__(cls, *args, **kwargs):
        if cls is BaseSerializer:
            raise TypeError("BaseSerializer cannot be instantiated")
        return super().__new__(cls)


def serializer_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseSerializer)



# Standard Library Imports
from collections import defaultdict
from enum import Enum

# Protean
from protean.core.entity import BaseEntity
from protean.core.field.basic import Auto, Integer, String


class AbstractPerson(BaseEntity):
    age = Integer(default=5)

    class Meta:
        abstract = True


class ConcretePerson(BaseEntity):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50)


class AdultAbstractPerson(ConcretePerson):
    age = Integer(default=21)

    class Meta:
        abstract = True


class Person(BaseEntity):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50)
    age = Integer(default=21)


class PersonAutoSSN(BaseEntity):
    ssn = Auto(identifier=True)
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50)
    age = Integer(default=21)


class PersonExplicitID(BaseEntity):
    ssn = String(max_length=36, identifier=True)
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50)
    age = Integer(default=21)


class Adult(Person):
    pass

    class Meta:
        schema_name = "adults"


class NotAPerson(BaseEntity):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50)
    age = Integer(default=21)


# Entities to test Meta Info overriding # START #
class DbPerson(BaseEntity):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50)
    age = Integer(default=21)

    class Meta:
        schema_name = "pepes"


class SqlPerson(Person):
    class Meta:
        schema_name = "people"


class DifferentDbPerson(Person):
    class Meta:
        provider = "non-default"


class SqlDifferentDbPerson(Person):
    class Meta:
        provider = "non-default-sql"


class OrderedPerson(BaseEntity):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50)
    age = Integer(default=21)

    class Meta:
        order_by = "first_name"


class OrderedPersonSubclass(Person):
    class Meta:
        order_by = "last_name"


class BuildingStatus(Enum):
    WIP = "WIP"
    DONE = "DONE"


class Building(BaseEntity):
    name = String(max_length=50)
    floors = Integer()
    status = String(choices=BuildingStatus)

    def defaults(self):
        if not self.status:
            if self.floors == 4:
                self.status = BuildingStatus.DONE.value
            else:
                self.status = BuildingStatus.WIP.value

    def clean(self):
        errors = defaultdict(list)

        if self.floors >= 4 and self.status != BuildingStatus.DONE.value:
            errors["status"].append("should be DONE")

        return errors



"""Module for defining different validators used by Field Types"""
# Standard Library Imports
import re

# Protean
from protean.core.exceptions import ValidationError


class MinLengthValidator:
    """ Validate the minimum length for the field value"""

    def __init__(self, min_length):
        self.min_length = min_length
        self.message = f"value has less than " f"{self.min_length} characters"

    def __call__(self, value):
        if self.min_length and len(value) < self.min_length:
            raise ValidationError(self.message)


class MaxLengthValidator:
    """ Validate the maximum length for the field value"""

    def __init__(self, max_length):
        self.max_length = max_length
        self.message = f"value has more than " f"{self.max_length} characters"

    def __call__(self, value):
        if self.max_length and len(value) > self.max_length:
            raise ValidationError(self.message)


class MinValueValidator:
    """ Validate the minimum value for the field"""

    def __init__(self, min_value):
        self.min_value = min_value
        self.message = f"value is lesser than {self.min_value}"

    def __call__(self, value):
        if self.min_value and value < self.min_value:
            raise ValidationError(self.message)


class MaxValueValidator:
    """ Validate the maximum value for the field"""

    def __init__(self, max_value):
        self.max_value = max_value
        self.message = f"value is greater than {self.max_value}"

    def __call__(self, value):
        if self.max_value and value > self.max_value:
            raise ValidationError(self.message)


class RegexValidator:
    """Validate the regex against given value"""

    regex = ""
    message = "invalid value"
    code = "invalid"
    inverse_match = False
    flags = 0

    def __init__(
        self, regex=None, message=None, code=None, inverse_match=None, flags=None
    ):
        if regex is not None:
            self.regex = regex
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code
        if inverse_match is not None:
            self.inverse_match = inverse_match
        if flags is not None:
            self.flags = flags
        if self.flags and not isinstance(self.regex, str):
            raise TypeError(
                "If the flags are set, regex must be a regular expression string."
            )

        self.regex = re.compile(self.regex, self.flags)

    def __call__(self, value):
        """
        Validate that the input contains (or does *not* contain, if
        inverse_match is True) a match for the regular expression.
        """
        regex_matches = self.regex.search(str(value))
        invalid_input = regex_matches if self.inverse_match else not regex_matches
        if invalid_input:
            raise ValidationError(self.message)



# Standard Library Imports
from datetime import datetime, timedelta
from uuid import uuid4

# Protean
import pytest

from protean.core.exceptions import (
    ObjectNotFoundError,
    TooManyObjectsError,
    ValidationError,
)
from protean.core.queryset import Q, QuerySet
from sqlalchemy.orm.session import Session

# Local/Relative Imports
from .elements import Person, User


@pytest.mark.postgresql
class TestDAO:
    """This class holds tests for DAO class"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    def test_successful_initialization_of_dao(self, test_domain):
        test_domain.get_dao(Person).query.all()
        provider = test_domain.get_provider("default")
        conn = provider.get_connection()
        assert conn is not None
        assert isinstance(conn, Session)

    def test_that_escaped_quotes_in_values_are_handled_properly(self, test_domain):
        test_domain.get_dao(Person).create(
            first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            first_name="Aramis", last_name="Musketeer", age=4
        )

        person1 = test_domain.get_dao(Person).create(
            first_name="d'Artagnan1", last_name="John", age=5
        )
        person2 = test_domain.get_dao(Person).create(
            first_name="d'Artagnan2", last_name="John", age=5
        )
        person3 = test_domain.get_dao(Person).create(
            first_name='d"Artagnan3', last_name="John", age=5
        )
        person4 = test_domain.get_dao(Person).create(
            first_name='d"Artagnan4', last_name="John", age=5
        )

        assert all(
            person is not None for person in [person1, person2, person3, person4]
        )


@pytest.mark.postgresql
class TestDAODeleteFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    def test_delete_an_object_in_repository_by_id(self, test_domain):
        """ Delete an object in the repository by ID"""
        identifier = uuid4()
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="John", last_name="Doe", age=22
        )

        persisted_person = test_domain.get_dao(Person).get(identifier)
        assert persisted_person is not None

        deleted_person = test_domain.get_dao(Person).delete(person)
        assert deleted_person is not None
        assert deleted_person.state_.is_destroyed is True

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(identifier)

    def test_delete_all_records_in_repository(self, test_domain):
        """Delete all objects in a repository"""
        test_domain.get_dao(Person).create(
            first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            first_name="dArtagnan", last_name="Musketeer", age=5
        )

        person_records = test_domain.get_dao(Person).query.filter(Q())
        assert person_records.total == 4

        test_domain.get_dao(Person).delete_all()

        person_records = test_domain.get_dao(Person).query.filter(Q())
        assert person_records.total == 0

    def test_deleting_a_persisted_entity(self, test_domain):
        """ Delete an object in the repository by ID"""
        identifier = uuid4()
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="Jim", last_name="Carrey"
        )
        deleted_person = test_domain.get_dao(Person).delete(person)
        assert deleted_person is not None
        assert deleted_person.state_.is_destroyed is True

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(identifier)

    def test_deleting_all_entities_of_a_type(self, test_domain):
        test_domain.get_dao(Person).create(
            first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            first_name="dArtagnan", last_name="Musketeer", age=5
        )

        people = test_domain.get_dao(Person).query.all()
        assert people.total == 4

        test_domain.get_dao(Person).delete_all()

        people = test_domain.get_dao(Person).query.all()
        assert people.total == 0

    def test_deleting_all_records_of_a_type_satisfying_a_filter(self, test_domain):
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        identifier4 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=identifier4, first_name="d'Artagnan", last_name="Musketeer", age=5
        )

        # Perform update
        deleted_count = test_domain.get_dao(Person).query.filter(age__gt=3).delete_all()

        # Query and check if only the relevant records have been deleted
        assert deleted_count == 2

        person1 = test_domain.get_dao(Person).get(identifier1)
        person2 = test_domain.get_dao(Person).get(identifier2)

        assert person1 is not None
        assert person2 is not None

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(identifier3)

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(identifier4)

    def test_deleting_records_satisfying_a_filter(self, test_domain):
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        identifier4 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=identifier4, first_name="d'Artagnan", last_name="Musketeer", age=5
        )

        # Perform update
        deleted_count = test_domain.get_dao(Person).query.filter(age__gt=3).delete()

        # Query and check if only the relevant records have been updated
        assert deleted_count == 2
        assert test_domain.get_dao(Person).query.all().total == 2

        assert test_domain.get_dao(Person).get(identifier1) is not None
        assert test_domain.get_dao(Person).get(identifier2) is not None
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(identifier3)

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(identifier4)


@pytest.mark.postgresql
class TestDAORetrievalFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    @pytest.fixture
    def identifier(self):
        return uuid4()

    @pytest.fixture
    def persisted_person(self, test_domain, identifier):
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="John", last_name="Doe"
        )
        return person

    def test_filtering_of_database_records(self, test_domain):
        # Add multiple entries to the database
        test_domain.get_dao(Person).create(first_name="John", last_name="Doe", age=22)
        test_domain.get_dao(Person).create(first_name="Jane", last_name="Doe", age=18)
        test_domain.get_dao(Person).create(first_name="Baby", last_name="Roe", age=2)

        # Filter by the last name
        people = test_domain.get_dao(Person).query.filter(last_name="Doe")
        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        # Order the results by age
        people = test_domain.get_dao(Person).query.order_by("age")
        assert people is not None
        assert people.first.age == 2
        assert people.first.first_name == "Baby"

    def test_traversal_of_filter_results(self, test_domain):
        """ Test the traversal of the filter results"""
        identifiers = [uuid4() for i in range(1, 5)]
        timestamp = datetime.now()
        for counter in range(1, 5):
            test_domain.get_dao(Person).create(
                id=identifiers[counter - 1],
                first_name=f"John{counter}",
                last_name="Doe",
                created_at=timestamp + timedelta(minutes=counter),
            )

        people = test_domain.get_dao(Person).query.limit(2).order_by("created_at").all()
        assert people is not None
        assert people.total == 4
        assert len(people.items) == 2
        assert people.first.id == identifiers[0]
        assert people.has_next
        assert not people.has_prev

        people = (
            test_domain.get_dao(Person)
            .query.offset(2)
            .limit(2)
            .order_by("created_at")
            .all()
        )
        assert len(people.items) == 2
        assert people.first.id == identifiers[2]
        assert not people.has_next
        assert people.has_prev

    def test_entity_retrieval_by_its_primary_key(
        self, test_domain, identifier, persisted_person
    ):
        """Test Entity Retrieval by its primary key"""
        person = test_domain.get_dao(Person).get(persisted_person.id)
        assert person is not None
        assert person.id == identifier

    def test_failed_entity_retrieval_by_its_primary_key(self, test_domain):
        """Test failed Entity Retrieval by its primary key"""
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).get(uuid4())

    def test_entity_retrieval_by_specific_column_value(
        self, test_domain, identifier, persisted_person
    ):
        person = test_domain.get_dao(Person).find_by(first_name="John")
        assert person is not None
        assert person.id == identifier

    def test_failed_entity_retrieval_by_column_value(
        self, test_domain, persisted_person
    ):
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).find_by(first_name="JohnnyChase")

    def test_entity_retrieval_by_multiple_column_values(self, test_domain):
        identifier1 = uuid4()
        identifier2 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Johnny1", last_name="Bravo", age=8
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Johnny2", last_name="Bravo", age=6
        )

        dog = test_domain.get_dao(Person).find_by(first_name="Johnny1", age=8)
        assert dog is not None
        assert dog.id == identifier1

    def test_failed_entity_retrieval_by_multiple_column_values(self, test_domain):
        test_domain.get_dao(Person).create(
            first_name="Johnny1", last_name="Bravo", age=8
        )
        test_domain.get_dao(Person).create(
            first_name="Johnny2", last_name="Bravo", age=6
        )

        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).find_by(first_name="Johnny1", age=6)

    def test_error_on_finding_multiple_results(self, test_domain):
        test_domain.get_dao(Person).create(
            first_name="Johnny1", last_name="Bravo", age=8
        )
        test_domain.get_dao(Person).create(
            first_name="Johnny1", last_name="Gravo", age=6
        )

        with pytest.raises(TooManyObjectsError):
            test_domain.get_dao(Person).find_by(first_name="Johnny1")

    def test_entity_query_initialization(self, test_domain):
        """Test the initialization of a QuerySet"""
        dao = test_domain.get_dao(Person)
        query = dao.query

        assert query is not None
        assert isinstance(query, QuerySet)
        assert query._criteria == QuerySet(dao, test_domain, Person)._criteria

    def test_filter_initialization_of_query_from_entity(self, test_domain):
        filters = [
            test_domain.get_dao(Person).query.filter(first_name="Murdock"),
            test_domain.get_dao(Person)
            .query.filter(first_name="Jean")
            .filter(last_name="John"),
            test_domain.get_dao(Person).query.offset(1),
            test_domain.get_dao(Person).query.limit(25),
            test_domain.get_dao(Person).query.order_by("first_name"),
            test_domain.get_dao(Person).query.exclude(first_name="Bravo"),
        ]

        for filter in filters:
            assert isinstance(filter, QuerySet)

    def test_that_chaining_filters_returns_a_queryset_for_further_chaining(
        self, test_domain
    ):
        person_query = test_domain.get_dao(Person).query.filter(name="Murdock")
        filters = [
            person_query,
            test_domain.get_dao(Person)
            .query.filter(first_name="Jean")
            .filter(last_name="John"),
            person_query.offset(5 * 5),
            person_query.limit(5),
            person_query.order_by("first_name"),
            person_query.exclude(last_name="Murdock"),
        ]

        for filter in filters:
            assert isinstance(filter, QuerySet)

    def test_filter_by_chaining_example_1(self, test_domain):
        # Add multiple entries to the DB
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Person attributes
        query = (
            test_domain.get_dao(Person)
            .query.filter(first_name="Jean")
            .filter(last_name="John")
            .filter(age=3)
        )
        people = query.all()

        assert people is not None
        assert people.total == 1
        assert len(people.items) == 1

        person = people.first
        assert person.id == identifier2

    def test_filter_by_chaining_example_2(self, test_domain):
        # Add multiple entries to the DB
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Person attributes
        query = test_domain.get_dao(Person).query.filter(last_name="John")
        people = query.all()

        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        person = people.first
        assert person.id == identifier1

    def test_filter_by_chaining_example_3(self, test_domain):
        # Add multiple entries to the DB
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        people = query.all()

        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        person = people.first
        assert person.id == identifier2

    def test_results_retrieved_with_filter(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(first_name="Jean", age=3, last_name="John")
        test_domain.get_dao(Person).create(first_name="Bart", age=6, last_name="Carrie")

        # Filter by the LastName
        people = test_domain.get_dao(Person).query.filter(last_name="John")
        assert people is not None
        assert people.total == 2
        assert len(people.items) == 2

        # Order the results by age
        people = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("-age")
        )
        assert people is not None
        assert people.first.age == 7
        assert people.first.first_name == "Murdock"

    def test_results_retrieved_after_exclusion(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(first_name="Jean", age=3, last_name="John")
        test_domain.get_dao(Person).create(first_name="Bart", age=6, last_name="Carrie")

        # Filter by Exclusion
        dogs = test_domain.get_dao(Person).query.exclude(last_name="John")
        assert dogs is not None
        assert dogs.total == 1
        assert len(dogs.items) == 1
        assert dogs.first.age == 6
        assert dogs.first.first_name == "Bart"

    def test_results_retrieved_after_multiple_value_exclusion(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(first_name="Jean", age=3, last_name="John")
        test_domain.get_dao(Person).create(first_name="Bart", age=6, last_name="Carrie")

        # Filter by the first_name
        people = test_domain.get_dao(Person).query.exclude(
            first_name__in=["Murdock", "Jean"]
        )
        assert people is not None
        assert people.total == 1
        assert len(people.items) == 1
        assert people.first.age == 6
        assert people.first.first_name == "Bart"

    def test_comparisons_using_different_operators(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(first_name="Jean", age=3, last_name="john")
        test_domain.get_dao(Person).create(first_name="Bart", age=6, last_name="Carrie")

        # Filter by the Owner
        people_gte = test_domain.get_dao(Person).query.filter(age__gte=3)
        people_datetime_gte = test_domain.get_dao(Person).query.filter(
            created_at__gte=datetime.now() - timedelta(hours=24)
        )
        people_lte = test_domain.get_dao(Person).query.filter(age__lte=6)
        people_gt = test_domain.get_dao(Person).query.filter(age__gt=3)
        people_lt = test_domain.get_dao(Person).query.filter(age__lt=6)
        people_in = test_domain.get_dao(Person).query.filter(
            first_name__in=["Jean", "Bart", "Nobody"]
        )
        people_exact = test_domain.get_dao(Person).query.filter(last_name__exact="John")
        people_iexact = test_domain.get_dao(Person).query.filter(
            last_name__iexact="John"
        )

        assert people_gte.total == 3
        assert people_datetime_gte.total == 3
        assert people_lte.total == 2
        assert people_gt.total == 2
        assert people_lt.total == 1
        assert people_in.total == 2
        assert people_exact.total == 1
        assert people_iexact.total == 2

    def test_filtering_using_contains(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(first_name="Jean", age=3, last_name="john")
        test_domain.get_dao(Person).create(first_name="Bart", age=6, last_name="Carrie")

        people_contains = test_domain.get_dao(Person).query.filter(
            last_name__contains="Joh"
        )
        people_icontains = test_domain.get_dao(Person).query.filter(
            last_name__icontains="Joh"
        )

        assert people_contains.total == 1
        assert people_icontains.total == 2

    def test_exception_on_usage_of_unsupported_comparison_operator(self, test_domain):
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(first_name="Jean", age=3, last_name="John")
        test_domain.get_dao(Person).create(first_name="Bart", age=6, last_name="Carrie")

        # Filter by the Owner
        with pytest.raises(NotImplementedError):
            test_domain.get_dao(Person).query.filter(age__notexact=3).all()


@pytest.mark.postgresql
class TestDAOSaveFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    def test_creation_throws_error_on_missing_fields(self, test_domain):
        """ Add an entity to the repository missing a required attribute"""
        with pytest.raises(ValidationError) as err:
            test_domain.get_dao(Person).create(last_name="Doe")

        assert err.value.messages == {"first_name": ["is required"]}

    def test_entity_persistence_with_create_method_and_fetch(self, test_domain):
        identifier = uuid4()
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="John", last_name="Doe"
        )
        assert person is not None
        assert person.first_name == "John"
        assert person.last_name == "Doe"
        assert person.age == 21

        db_person = test_domain.get_dao(Person).get(identifier)
        assert db_person is not None
        assert db_person == person

    def test_multiple_persistence_for_an_aggregate(self, test_domain):
        """Test that save can be invoked again on an already existing entity, to update values"""
        person = Person(first_name="Johnny", last_name="John")
        test_domain.get_dao(Person).save(person)

        person.last_name = "Janey"
        test_domain.get_dao(Person).save(person)

        test_domain.get_dao(Person).get(person.id)
        assert person.last_name == "Janey"


@pytest.mark.postgresql
class TestDAOUpdateFunctionality:
    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    def test_update_an_existing_entity_in_the_repository(self, test_domain):
        identifier = uuid4()
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="John", last_name="Doe", age=22
        )

        test_domain.get_dao(Person).update(person, age=10)
        updated_person = test_domain.get_dao(Person).get(identifier)
        assert updated_person is not None
        assert updated_person.age == 10

    def test_that_updating_a_deleted_aggregate_raises_object_not_found_error(
        self, test_domain
    ):
        """Try to update a non-existing entry"""

        person = test_domain.get_dao(Person).create(
            first_name="Johnny", last_name="John"
        )
        test_domain.get_dao(Person).delete(person)
        with pytest.raises(ObjectNotFoundError):
            test_domain.get_dao(Person).update(person, {"age": 10})

    def test_updating_record_with_dictionary_args(self, test_domain):
        """ Update an existing entity in the repository"""
        identifier = uuid4()
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="Johnny", last_name="John", age=2
        )

        test_domain.get_dao(Person).update(person, {"age": 10})
        u_person = test_domain.get_dao(Person).get(identifier)
        assert u_person is not None
        assert u_person.age == 10

    def test_updating_record_with_kwargs(self, test_domain):
        """ Update an existing entity in the repository"""
        identifier = uuid4()
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="Johnny", last_name="John", age=2
        )

        test_domain.get_dao(Person).update(person, age=10)
        u_person = test_domain.get_dao(Person).get(identifier)
        assert u_person is not None
        assert u_person.age == 10

    def test_updating_record_with_both_dictionary_args_and_kwargs(self, test_domain):
        """ Update an existing entity in the repository"""
        identifier = uuid4()
        person = test_domain.get_dao(Person).create(
            id=identifier, first_name="Johnny", last_name="John", age=2
        )

        test_domain.get_dao(Person).update(person, {"first_name": "Stephen"}, age=10)
        u_person = test_domain.get_dao(Person).get(identifier)
        assert u_person is not None
        assert u_person.age == 10
        assert u_person.first_name == "Stephen"

    def test_updating_record_through_filter(self, test_domain):
        """Test that update by query updates only correct records"""
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        identifier4 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=identifier4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        # Perform update
        updated_count = (
            test_domain.get_dao(Person)
            .query.filter(age__gt=3)
            .update(last_name="Fraud")
        )

        # Query and check if only the relevant records have been updated
        assert updated_count == 2

        u_person1 = test_domain.get_dao(Person).get(identifier1)
        u_person2 = test_domain.get_dao(Person).get(identifier2)
        u_person3 = test_domain.get_dao(Person).get(identifier3)
        u_person4 = test_domain.get_dao(Person).get(identifier4)
        assert u_person1.last_name == "Musketeer"
        assert u_person2.last_name == "Musketeer"
        assert u_person3.last_name == "Fraud"
        assert u_person4.last_name == "Fraud"

    def test_updating_multiple_records_through_filter_with_arg_value(self, test_domain):
        """Try updating all records satisfying filter in one step, passing a dict"""
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        identifier4 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=identifier4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        # Perform update
        updated_count = (
            test_domain.get_dao(Person)
            .query.filter(age__gt=3)
            .update_all({"last_name": "Fraud"})
        )

        # Query and check if only the relevant records have been updated
        assert updated_count == 2

        u_person1 = test_domain.get_dao(Person).get(identifier1)
        u_person2 = test_domain.get_dao(Person).get(identifier2)
        u_person3 = test_domain.get_dao(Person).get(identifier3)
        u_person4 = test_domain.get_dao(Person).get(identifier4)
        assert u_person1.last_name == "Musketeer"
        assert u_person2.last_name == "Musketeer"
        assert u_person3.last_name == "Fraud"
        assert u_person4.last_name == "Fraud"

    def test_updating_multiple_records_through_filter_with_kwarg_value(
        self, test_domain
    ):
        """Try updating all records satisfying filter in one step"""
        identifier1 = uuid4()
        identifier2 = uuid4()
        identifier3 = uuid4()
        identifier4 = uuid4()
        test_domain.get_dao(Person).create(
            id=identifier1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=identifier2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=identifier3, first_name="Aramis", last_name="Musketeer", age=4
        )
        test_domain.get_dao(Person).create(
            id=identifier4, first_name="dArtagnan", last_name="Musketeer", age=5
        )

        # Perform update
        updated_count = (
            test_domain.get_dao(Person)
            .query.filter(age__gt=3)
            .update_all(last_name="Fraud")
        )

        # Query and check if only the relevant records have been updated
        assert updated_count == 2

        u_person1 = test_domain.get_dao(Person).get(identifier1)
        u_person2 = test_domain.get_dao(Person).get(identifier2)
        u_person3 = test_domain.get_dao(Person).get(identifier3)
        u_person4 = test_domain.get_dao(Person).get(identifier4)
        assert u_person1.last_name == "Musketeer"
        assert u_person2.last_name == "Musketeer"
        assert u_person3.last_name == "Fraud"
        assert u_person4.last_name == "Fraud"


@pytest.mark.postgresql
class TestDAOValidations:
    """This class holds tests for DAO class"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(User)

    def test_unique(self, test_domain):
        """ Test the unique constraints for the entity """
        test_domain.get_dao(User).create(email="john.doe@gmail.com", password="a1b2c3")

        with pytest.raises(ValidationError) as err:
            test_domain.get_dao(User).create(
                email="john.doe@gmail.com", password="d4e5r6"
            )
        assert err.value.messages == {
            "email": ["`User` with this `email` already exists."]
        }

    def test_that_field_required_validations_are_triggered_on_incorrect_attribute_setting(
        self,
    ):
        """Test failed `save()` because of validation errors"""
        person = Person(first_name="Johnny", last_name="John")

        with pytest.raises(ValidationError) as error:
            person.first_name = ""  # Simulate an error by force-resetting an attribute

        assert error.value.messages == {"first_name": ["is required"]}

    def test_that_primitive_validations_on_type_are_thrown_correctly_on_initialization(
        self, test_domain
    ):
        with pytest.raises(ValidationError) as error:
            Person(first_name="Johnny", last_name="John", age="x")

        assert error.value.messages == {"age": ['"x" value must be an integer.']}

    def test_that_primitive_validations_on_type_are_thrown_correctly_on_update(
        self, test_domain
    ):
        person = test_domain.get_dao(Person).create(
            first_name="John", last_name="Doe", age=22
        )

        with pytest.raises(ValidationError) as error:
            test_domain.get_dao(Person).update(
                person, age="x"
            )  # Age should be an integer

        assert error.value.messages == {"age": ['"x" value must be an integer.']}


@pytest.mark.postgresql
class TestDAOLookup:
    """This class holds tests for Lookup Class"""

    @pytest.fixture
    def sample_lookup_cls(self):
        # Protean
        from protean.adapters.repository.sqlalchemy import SAProvider
        from protean.port.dao import BaseLookup

        @SAProvider.register_lookup
        class SampleLookup(BaseLookup):
            """A simple implementation of lookup class"""

            lookup_name = "sample"

            def as_expression(self):
                return "%s %s %s" % (
                    self.process_source(),
                    "<<<>>>",
                    self.process_target(),
                )

        return SampleLookup

    def test_initialization_of_a_lookup_object(self, sample_lookup_cls):
        lookup = sample_lookup_cls("src", "trg")
        assert lookup.as_expression() == "src <<<>>> trg"

    def test_registration_of_a_lookup_to_an_adapter(self, sample_lookup_cls):
        # Protean
        from protean.adapters.repository.sqlalchemy import SAProvider

        assert SAProvider.get_lookups().get("sample") == sample_lookup_cls



"""Value Object Functionality and Classes"""
# Standard Library Imports
import logging

# Protean
from protean.core.exceptions import IncorrectUsageError
from protean.utils import DomainObjects, derive_element_class
from protean.utils.container import BaseContainer

logger = logging.getLogger("protean.domain.value_object")


class BaseValueObject(BaseContainer):
    """The Base class for Protean-Compliant Domain Value Objects.

    Provides helper methods to custom define attributes, and find attribute names
    during runtime.

    Basic Usage::

        @ValueObject
        class Address:
            unit = field.String()
            address = field.String(required=True, max_length=255)
            city = field.String(max_length=50)
            province = field.String(max_length=2)
            pincode = field.String(max_length=6)

    (or)

        class Address(BaseValueObject):
            unit = field.String()
            address = field.String(required=True, max_length=255)
            city = field.String(max_length=50)
            province = field.String(max_length=2)
            pincode = field.String(max_length=6)

        domain.register_element(Address)

    If persistence is required, the model associated with this value object is retrieved dynamically.
    The value object may be persisted along with its related entity, or separately in which case its model is
    retrieved from the repository factory. Model is usually initialized with a live DB connection.
    """

    element_type = DomainObjects.VALUE_OBJECT

    def __new__(cls, *args, **kwargs):
        if cls is BaseValueObject:
            raise TypeError("BaseValueObject cannot be instantiated")
        return super().__new__(cls)


def value_object_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseValueObject)

    if not (
        hasattr(element_cls.meta_, "aggregate_cls") and element_cls.meta_.aggregate_cls
    ):
        element_cls.meta_.aggregate_cls = kwargs.pop("aggregate_cls", None)

    if not element_cls.meta_.aggregate_cls:
        raise IncorrectUsageError(
            "Value Objects need to be associated with an Aggregate"
        )

    return element_cls



"""This module implements the central domain object, along with decorators
to register Domain Elements.
"""
# Standard Library Imports
import logging
import sys

# Protean
from protean.adapters import Brokers, EmailProviders, Providers
from protean.core.exceptions import (
    IncorrectUsageError,
    NotSupportedError,
    ObjectNotFoundError,
)
from protean.domain.registry import _DomainRegistry
from protean.utils import DomainObjects, fully_qualified_name
from werkzeug.datastructures import ImmutableDict

# Local/Relative Imports
from .config import Config, ConfigAttribute
from .context import DomainContext, _DomainContextGlobals
from .helpers import _PackageBoundObject, get_debug_flag, get_env

logger = logging.getLogger("protean.domain")

# a singleton sentinel value for parameter defaults
_sentinel = object()


class Domain(_PackageBoundObject):
    """The domain object is a one-stop gateway to:
    * Registering Domain Objects/Concepts
    * Querying/Retrieving Domain Artifacts like Entities, Services, etc.
    * Retrieve injected infrastructure adapters

    Usually you create a :class:`Domain` instance in your main module or
    in the :file:`__init__.py` file of your package like this::

        from protean import Domain
        domain = Domain(__name__)

    :param domain_name: the name of the domain
    """

    # Protean
    from protean.utils import IdentityStrategy, IdentityType

    config_class = Config
    domain_context_globals_class = _DomainContextGlobals

    #: What environment the app is running in Protean and extensions may
    #: enable behaviors based on the environment, such as enabling debug
    #: mode. This maps to the :data:`ENV` config key. This is set by the
    #: :envvar:`PROTEAN_ENV` environment variable and may not behave as
    #: expected if set in code.
    #:
    #: **Do not enable development when deploying in production.**
    #:
    #: Default: ``'production'``
    env = ConfigAttribute("ENV")

    #: The testing flag.  Set this to ``True`` to enable the test mode of
    #: Protean extensions (and in the future probably also Protean itself).
    #: For example this might activate test helpers that have an
    #: additional runtime cost which should not be enabled by default.
    #:
    #: This attribute can also be configured from the config with the
    #: ``TESTING`` configuration key.  Defaults to ``False``.
    testing = ConfigAttribute("TESTING")

    #: If a secret key is set, cryptographic components can use this to
    #: sign cookies and other things. Set this to a complex random value
    #: when you want to use the secure cookie for instance.
    #:
    #: This attribute can also be configured from the config with the
    #: :data:`SECRET_KEY` configuration key. Defaults to ``None``.
    secret_key = ConfigAttribute("SECRET_KEY")

    root_path = None

    default_config = ImmutableDict(
        {
            "ENV": None,
            "DEBUG": None,
            "SECRET_KEY": None,
            "AUTOLOAD_DOMAIN": True,
            "IDENTITY_STRATEGY": IdentityStrategy.UUID,
            "IDENTITY_TYPE": IdentityType.STRING,
            "DATABASES": {"default": {"PROVIDER": "protean.adapters.MemoryProvider"}},
            "BROKERS": {"default": {"PROVIDER": "protean.adapters.InlineBroker"}},
            "EMAIL_PROVIDERS": {
                "default": {
                    "PROVIDER": "protean.adapters.DummyEmailProvider",
                    "DEFAULT_FROM_EMAIL": "admin@team8solutions.com",
                },
            },
            "AGGREGATE_CHILDREN_LIMIT": 100,
        },
    )

    def __init__(
        self, domain_name=__name__, root_path=None, instance_relative_config=False
    ):

        _PackageBoundObject.__init__(
            self, domain_name, root_path=root_path,
        )

        self.domain_name = domain_name

        # Registry for all domain Objects
        self._domain_registry = _DomainRegistry()

        #: The configuration dictionary as :class:`Config`.  This behaves
        #: exactly like a regular dictionary but supports additional methods
        #: to load a config from files.
        self.config = self.make_config(instance_relative_config)

        self.providers = Providers(self)
        self.brokers = Brokers(self)
        self.email_providers = EmailProviders(self)

        # Cache for holding Model to Entity/Aggregate associations
        self._models = {}
        self._constructed_models = {}

        #: A list of functions that are called when the domain context
        #: is destroyed.  This is the place to store code that cleans up and
        #: disconnects from databases, for example.
        self.teardown_domain_context_functions = []

    def init(self):
        """ Parse the domain folder, and attach elements dynamically to the domain.

        Protean parses all files in the domain file's folder, as well as under it,
        to load elements. So, all domain files are to be nested under the file contain
        the domain definition.

        One can use the `AUTOLOAD_DOMAIN` flag in Protean config, `True` by default,
        to control this functionality.

        When enabled, Protean is responsible for loading domain elements and ensuring
        all functionality is activated.

        The developer is responsible for activating functionality manually when
        autoloading is disabled. Element activation can be done by importing them
        in central areas of domain execution, like Application Services.

        For example, asynchronous aspects of a domain like its Subscribers and
        Event Handlers should be imported in their relevant Application Services
        and Aggregates.

        This method bubbles up circular import issues, if present, in the domain code.
        """
        if self.config["AUTOLOAD_DOMAIN"] is True:
            # Standard Library Imports
            import importlib.util
            import inspect
            import os
            import pathlib

            # Fetch the domain file and derive the system path
            domain_path = inspect.stack()[1][
                1
            ]  # Find the file in which the domain is defined
            dir_name = pathlib.PurePath(pathlib.Path(domain_path).resolve()).parent
            path = pathlib.Path(dir_name)  # Resolve the domain file's directory
            system_folder_path = (
                path.parent
            )  # Get the directory of the domain file to traverse from

            logger.debug(f"Loading domain from {dir_name}...")

            for root, dirs, files in os.walk(dir_name):
                if pathlib.PurePath(root).name not in ["__pycache__"]:
                    package_path = root[len(str(system_folder_path)) + 1 :]
                    module_name = package_path.replace(os.sep, ".")

                    for file in files:
                        file_base_name = os.path.basename(file)

                        # Construct the module path to import from
                        if file_base_name != "__init__":
                            sub_module_name = os.path.splitext(file_base_name)[0]
                            file_module_name = module_name + "." + sub_module_name
                        else:
                            file_module_name = module_name
                        full_file_path = os.path.join(root, file)

                        try:
                            if (
                                full_file_path != domain_path
                            ):  # Don't load the domain file itself again
                                spec = importlib.util.spec_from_file_location(
                                    file_module_name, full_file_path
                                )
                                module = importlib.util.module_from_spec(spec)
                                spec.loader.exec_module(module)

                                logger.debug(f"Loaded {file_module_name}")
                        except ModuleNotFoundError as exc:
                            logger.error(f"Error while loading a module: {exc}")
                        except ModuleNotFoundError as exc:
                            logger.error(f"Error while autoloading modules: {exc}")

    def make_config(self, instance_relative=False):
        """Used to create the config attribute by the Domain constructor.
        The `instance_relative` parameter is passed in from the constructor
        of Domain (there named `instance_relative_config`) and indicates if
        the config should be relative to the instance path or the root path
        of the application.
        """
        root_path = self.root_path
        if instance_relative:
            root_path = self.instance_path
        defaults = dict(self.default_config)
        defaults["ENV"] = get_env()
        defaults["DEBUG"] = get_debug_flag()
        return self.config_class(root_path, defaults)

    def domain_context(self):
        """Create an :class:`~protean.context.DomainContext`. Use as a ``with``
        block to push the context, which will make :data:`current_domain`
        point at this domain.

        ::

            with domain.domain_context():
                init_db()
        """
        return DomainContext(self)

    def teardown_domain_context(self, f):
        """Registers a function to be called when the domain context
        ends.

        Example::

            ctx = domain.domain_context()
            ctx.push()
            ...
            ctx.pop()

        When ``ctx.pop()`` is executed in the above example, the teardown
        functions are called just before the domain context moves from the
        stack of active contexts.  This becomes relevant if you are using
        such constructs in tests.

        When a teardown function was called because of an unhandled exception
        it will be passed an error object. If an :meth:`errorhandler` is
        registered, it will handle the exception and the teardown will not
        receive it.

        The return values of teardown functions are ignored.
        """
        self.teardown_domain_context_functions.append(f)
        return f

    def do_teardown_domain_context(self, exc=_sentinel):
        """Called right before the domain context is popped.

        This calls all functions decorated with
        :meth:`teardown_domain_context`.

        This is called by
        :meth:`DomainContext.pop() <protean.context.DomainContext.pop>`.
        """
        if exc is _sentinel:
            exc = sys.exc_info()[1]
        for func in reversed(self.teardown_domain_context_functions):
            func(exc)

    @property
    def registry(self):
        return self._domain_registry

    def factory_for(self, domain_object_type):
        # Protean
        from protean.core.aggregate import aggregate_factory
        from protean.core.application_service import application_service_factory
        from protean.core.command import command_factory
        from protean.core.command_handler import command_handler_factory
        from protean.core.domain_event import domain_event_factory
        from protean.core.domain_service import domain_service_factory
        from protean.core.email import email_factory
        from protean.core.entity import entity_factory
        from protean.core.model import model_factory
        from protean.core.repository import repository_factory
        from protean.core.serializer import serializer_factory
        from protean.core.subscriber import subscriber_factory
        from protean.core.value_object import value_object_factory
        from protean.core.view import view_factory

        factories = {
            DomainObjects.AGGREGATE.value: aggregate_factory,
            DomainObjects.APPLICATION_SERVICE.value: application_service_factory,
            DomainObjects.COMMAND.value: command_factory,
            DomainObjects.COMMAND_HANDLER.value: command_handler_factory,
            DomainObjects.DOMAIN_EVENT.value: domain_event_factory,
            DomainObjects.DOMAIN_SERVICE.value: domain_service_factory,
            DomainObjects.EMAIL.value: email_factory,
            DomainObjects.ENTITY.value: entity_factory,
            DomainObjects.MODEL.value: model_factory,
            DomainObjects.REPOSITORY.value: repository_factory,
            DomainObjects.SUBSCRIBER.value: subscriber_factory,
            DomainObjects.SERIALIZER.value: serializer_factory,
            DomainObjects.VALUE_OBJECT.value: value_object_factory,
            DomainObjects.VIEW.value: view_factory,
        }

        if domain_object_type.value not in factories:
            raise IncorrectUsageError(
                "Unknown Element Type {element_type.value} for class {element_cls.__name__} "
                " (Error: {exc})",
            )

        return factories[domain_object_type.value]

    def _register_element(self, element_type, element_cls, **kwargs):  # noqa: C901
        """Register class into the domain"""
        # Check if `element_cls` is already a subclass of the Element Type
        #   which would be the case in an explicit declaration like `class Account(BaseEntity):`
        #
        # We will need to construct a class derived from the right base class
        #   if the Element was specified through annotation, like so:
        #
        #  ```
        #       @Entity
        #       class Account:
        #  ```

        new_cls = self.factory_for(element_type)(element_cls, **kwargs)

        if element_type == DomainObjects.MODEL:
            # Remember model association with aggregate/entity class, for easy fetching
            self._models[fully_qualified_name(new_cls.meta_.entity_cls)] = new_cls

        # Register element with domain
        self._domain_registry.register_element(new_cls)

        return new_cls

    # _cls should never be specified by keyword, so start it with an
    # underscore.  The presence of _cls is used to detect if this
    # decorator is being called with parameters or not.
    def _domain_element(
        self, element_type, _cls=None, **kwargs,
    ):
        """Returns the registered class after decoarating it and recording its presence in the domain"""

        def wrap(cls):
            return self._register_element(element_type, cls, **kwargs)

        # See if we're being called as @Entity or @Entity().
        if _cls is None:
            # We're called with parens.
            return wrap

        # We're called as @dataclass without parens.
        return wrap(_cls)

    def register_model(self, model_cls, **kwargs):
        """Register a model class"""
        return self._register_element(DomainObjects.MODEL, model_cls, **kwargs)

    def register(self, element_cls, **kwargs):
        """Register an element already subclassed with the correct Hierarchy"""
        if getattr(element_cls, "element_type", None) not in [
            element for element in DomainObjects
        ]:
            raise NotImplementedError

        if (
            hasattr(element_cls, "meta_")
            and hasattr(element_cls.meta_, "abstract")
            and element_cls.meta_.abstract is True
        ):
            raise NotSupportedError(
                f"{element_cls.__name__} class has been marked abstract"
                " and cannot be instantiated"
            )

        return self._register_element(element_cls.element_type, element_cls, **kwargs)

    def delist(self, element_cls):
        """Delist a Domain Element.

        This method will result in a no-op if the entity class was not found
        in the registry for whatever reason.
        """
        if getattr(element_cls, "element_type", None) not in [
            element for element in DomainObjects
        ]:
            raise NotImplementedError

        self._domain_registry.dei_element(element_cls.element_type, element_cls)

    def _get_element_by_name(self, element_types, element_name):
        """Fetch Domain record with an Element name"""
        for element_type in element_types:
            if element_name in self._domain_registry._elements[element_type.value]:
                return self._domain_registry._elements[element_type.value][element_name]
        else:
            raise ObjectNotFoundError(
                f"Element {element_name} not registered in domain {self.domain_name}"
            )

    def _get_element_by_class(self, element_types, element_cls):
        """Fetch Domain record with Element class details"""
        element_qualname = fully_qualified_name(element_cls)
        return self._get_element_by_name(element_types, element_qualname)

    def _replace_element_by_class(self, new_element_cls):
        aggregate_record = self._get_element_by_class(
            (DomainObjects.AGGREGATE, DomainObjects.ENTITY), new_element_cls
        )

        self._domain_registry._elements[aggregate_record.class_type][
            aggregate_record.qualname
        ].cls = new_element_cls

    ######################
    # Element Decorators #
    ######################

    def aggregate(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.AGGREGATE, _cls=_cls, **kwargs,)

    def application_service(self, _cls=None, **kwargs):
        return self._domain_element(
            DomainObjects.APPLICATION_SERVICE, _cls=_cls, **kwargs,
        )

    def command(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.COMMAND, _cls=_cls, **kwargs,)

    def command_handler(self, command, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.COMMAND_HANDLER, _cls=_cls, **kwargs)

    def domain_event(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.DOMAIN_EVENT, _cls=_cls, **kwargs,)

    def domain_service(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.DOMAIN_SERVICE, _cls=_cls, **kwargs,)

    def entity(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.ENTITY, _cls=_cls, **kwargs)

    def email(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.EMAIL, _cls=_cls, **kwargs)

    def model(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.MODEL, _cls=_cls, **kwargs)

    def repository(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.REPOSITORY, _cls=_cls, **kwargs)

    def serializer(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.SERIALIZER, _cls=_cls, **kwargs)

    def subscriber(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.SUBSCRIBER, _cls=_cls, **kwargs,)

    def value_object(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.VALUE_OBJECT, _cls=_cls, **kwargs,)

    def view(self, _cls=None, **kwargs):
        return self._domain_element(DomainObjects.VIEW, _cls=_cls, **kwargs,)

    ########################
    # Broker Functionality #
    ########################

    def publish(self, domain_event):
        self.brokers.publish(domain_event)

    def publish_command(self, command):
        self.brokers.publish_command(command)

    ############################
    # Repository Functionality #
    ############################

    def get_model(self, aggregate_cls):
        return self.providers.get_model(aggregate_cls)

    def has_provider(self, provider_name):
        return self.providers.has_provider(provider_name)

    def get_provider(self, provider_name):
        return self.providers.get_provider(provider_name)

    def get_connection(self, provider_name="default"):
        return self.providers.get_connection(provider_name)

    def providers_list(self):
        return self.providers.providers_list()

    def repository_for(self, aggregate_cls):
        return self.providers.repository_for(aggregate_cls)

    def get_dao(self, aggregate_cls):
        return self.providers.get_dao(aggregate_cls)

    #######################
    # Email Functionality #
    #######################

    def get_email_provider(self, provider_name):
        return self.email_providers.get_email_provider(provider_name)

    def send_email(self, email):
        return self.email_providers.send_email(email)



# Standard Library Imports
import os

# Protean
import pytest

from redis import Redis


def initialize_domain():
    # Protean
    from protean.domain import Domain

    domain = Domain("Celery Tests")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    domain.domain_context().push()
    return domain


@pytest.fixture(autouse=True)
def test_domain():
    domain = initialize_domain()

    yield domain


@pytest.fixture(scope="module")
def test_domain_for_worker():
    domain = initialize_domain()

    yield domain


@pytest.fixture(scope="session", autouse=True)
def setup_redis():
    # Initialize Redis
    # FIXME

    yield

    # Close connection to Redis
    # FIXME


@pytest.fixture(autouse=True)
def run_around_tests(test_domain):

    yield

    # Flush all in Redis
    # FIXME

    if test_domain.providers.has_provider("default"):
        test_domain.get_provider("default")._data_reset()



# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.application_service import BaseApplicationService
from protean.core.command import BaseCommand
from protean.core.domain_event import BaseDomainEvent
from protean.core.field.basic import Integer, String
from protean.core.field.embedded import AggregateField
from protean.core.unit_of_work import UnitOfWork
from protean.globals import current_domain


class PersonCommand(BaseCommand):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)

    @classmethod
    def add_newcomer(cls, person_dict):
        """Factory method to add a new Person to the system"""
        newcomer = Person(
            first_name=person_dict["first_name"],
            last_name=person_dict["last_name"],
            age=person_dict["age"],
        )

        # Publish Event via the domain
        current_domain.publish(PersonAdded(person=newcomer))

        return newcomer


class PersonAdded(BaseDomainEvent):
    person = AggregateField(Person)

    class Meta:
        aggregate_cls = Person


class PersonService(BaseApplicationService):
    @classmethod
    def add(cls, command: PersonCommand):
        with UnitOfWork():
            newcomer = Person.add_newcomer(command.to_dict())

            person_repo = current_domain.repository_for(Person)
            person_repo.add(newcomer)

        return newcomer



# cSpell: disable

# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# Standard Library Imports
import os

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]
if os.getenv("SPELLCHECK"):
    extensions += "sphinxcontrib.spelling"
    spelling_show_suggestions = True
    spelling_lang = "en_US"

source_suffix = ".rst"
master_doc = "index"
project = "Protean"
year = "2020"
author = "Subhash Bhushan C"
copyright = "{0}, {1}".format(year, author)
version = release = "0.5.6"

pygments_style = "autumn"
templates_path = ["."]
extlinks = {
    "issue": ("https://github.com/proteanhq/protean/issues/%s", "#"),
    "pr": ("https://github.com/proteanhq/protean/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
# on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

# if not on_rtd:  # only set the theme if we're building docs locally
#     html_theme = 'alabaster'

# Change theme to Alabaster
html_theme = "sphinx_rtd_theme"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

html_theme_options = {
    "style_nav_header_background": "#CF5400",
}
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]



# Protean
from protean.core.field.basic import String
from protean.utils.container import BaseContainer


class CustomBaseContainer(BaseContainer):
    def __new__(cls, *args, **kwargs):
        if cls is CustomBaseContainer:
            raise TypeError("CustomBaseContainer cannot be instantiated")
        return super().__new__(cls)


class CustomContainer(CustomBaseContainer):
    foo = String(max_length=50, required=True)
    bar = String(max_length=50, required=True)



# Protean
import pytest

from protean.core.exceptions import ValidationError

# Local/Relative Imports
from .elements import Person, PersonRepository, User


class TestValidations:
    """This class holds tests for DAO class"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)
        test_domain.register(User)

    def test_unique(self, test_domain):
        """ Test the unique constraints for the entity """
        test_domain.get_dao(User).create(email="john.doe@gmail.com", password="a1b2c3")

        with pytest.raises(ValidationError) as err:
            test_domain.get_dao(User).create(
                email="john.doe@gmail.com", password="d4e5r6"
            )
        assert err.value.messages == {
            "email": ["`User` with this `email` already exists."]
        }

    def test_that_field_required_validations_are_triggered_on_incorrect_attribute_setting(
        self,
    ):
        """Test failed `save()` because of validation errors"""
        person = Person(first_name="Johnny", last_name="John")

        with pytest.raises(ValidationError) as error:
            person.first_name = ""  # Simulate an error by force-resetting an attribute

        assert error.value.messages == {"first_name": ["is required"]}

    def test_that_primitive_validations_on_type_are_thrown_correctly_on_initialization(
        self, test_domain
    ):
        with pytest.raises(ValidationError) as error:
            Person(first_name="Johnny", last_name="John", age="x")

        assert error.value.messages == {"age": ['"x" value must be an integer.']}

    def test_that_primitive_validations_on_type_are_thrown_correctly_on_update(
        self, test_domain
    ):
        person = test_domain.get_dao(Person).create(
            first_name="John", last_name="Doe", age=22
        )

        with pytest.raises(ValidationError) as error:
            test_domain.get_dao(Person).update(
                person, age="x"
            )  # Age should be an integer

        assert error.value.messages == {"age": ['"x" value must be an integer.']}



# Standard Library Imports
import os

# Protean
import pytest


def initialize_domain():
    # Protean
    from protean.domain import Domain

    domain = Domain("SQLAlchemy Test - SQLite")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    return domain


@pytest.fixture(autouse=True)
def test_domain():
    domain = initialize_domain()

    with domain.domain_context():
        yield domain


@pytest.fixture(autouse=True)
def run_around_tests(test_domain):
    yield
    if test_domain.providers.has_provider("default"):
        test_domain.get_provider("default")._data_reset()



# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import Integer, String
from protean.core.repository import BaseRepository


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)


class PersonRepository(BaseRepository):
    pass



# Protean
import pytest

from protean.core.command import BaseCommand
from protean.core.exceptions import InvalidDataError, InvalidOperationError
from protean.core.field.basic import String
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import UserRegistrationCommand


class TestCommandInitialization:
    def test_that_command_object_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseCommand()

    def test_that_a_concrete_dto_can_be_instantiated(self):
        command = UserRegistrationCommand(
            email="john.doe@gmail.com", username="john.doe", password="secret1!",
        )
        assert command is not None

    def test_that_invalid_data_input_throws_an_exception(self):
        with pytest.raises(InvalidDataError) as exception1:
            UserRegistrationCommand(
                foo="bar", username="john.doe", password="secret1!",
            )
        assert exception1.value.messages == {"foo": ["is invalid"]}

        with pytest.raises(InvalidDataError) as exception2:
            UserRegistrationCommand(
                email="john.doe@gmail.com",
                username="123456789012345678901234567890123456789012345678901234567890",
                password="secret1!",
            )
        assert exception2.value.messages == {
            "username": ["value has more than 50 characters"]
        }


class TestCommandRegistration:
    def test_that_command_can_be_registered_with_domain(self, test_domain):
        test_domain.register(UserRegistrationCommand)

        assert (
            fully_qualified_name(UserRegistrationCommand)
            in test_domain.registry.commands
        )

    def test_that_command_can_be_registered_via_annotations(self, test_domain):
        @test_domain.command
        class ChangePasswordCommand:
            old_password = String(required=True, max_length=255)
            new_password = String(required=True, max_length=255)

        assert (
            fully_qualified_name(ChangePasswordCommand) in test_domain.registry.commands
        )


class TestCommandProperties:
    def test_two_commands_with_equal_values_are_considered_equal(self):
        command1 = UserRegistrationCommand(
            email="john.doe@gmail.com", username="john.doe", password="secret1!"
        )
        command2 = UserRegistrationCommand(
            email="john.doe@gmail.com", username="john.doe", password="secret1!"
        )

        assert command1 == command2

    @pytest.mark.xfail
    def test_that_commands_are_immutable(self):
        command = UserRegistrationCommand(
            email="john.doe@gmail.com", username="john.doe", password="secret1!"
        )
        with pytest.raises(InvalidOperationError):
            command.username = "jane.doe"

    def test_output_to_dict(self):
        command = UserRegistrationCommand(
            email="john.doe@gmail.com", username="john.doe", password="secret1!"
        )
        assert command.to_dict() == {
            "email": "john.doe@gmail.com",
            "username": "john.doe",
            "password": "secret1!",
        }

    def test_different_commands_are_distinct(self):
        command1 = UserRegistrationCommand(
            email="john.doe@gmail.com", username="john.doe", password="secret1!"
        )
        command2 = UserRegistrationCommand(
            email="jane.doe@gmail.com", username="jane.doe", password="not-so-secret!"
        )

        assert command1 != command2



# Standard Library Imports
import logging
import os
import pkgutil
import sys

logger = logging.getLogger("protean.application")


def get_env():
    """Get the environment the app is running in, indicated by the
    :envvar:`PROTEAN_ENV` environment variable. The default is
    ``'production'``.
    """
    return os.environ.get("PROTEAN_ENV") or "production"


def get_debug_flag():
    """Get whether debug mode should be enabled for the app, indicated
    by the :envvar:`PROTEAN_DEBUG` environment variable. The default is
    ``True`` if :func:`.get_env` returns ``'development'``, or ``False``
    otherwise.
    """
    val = os.environ.get("PROTEAN_DEBUG")

    if not val:
        return get_env() == "development"

    return val.lower() not in ("0", "false", "no")


def get_root_path(import_name):
    """Returns the path to a package or cwd if that cannot be found.  This
    returns the path of a package or the folder that contains a module.

    Not to be confused with the package path returned by :func:`find_package`.
    """
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, "__file__"):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == "__main__":
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, "get_filename"):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, "__file__", None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError(
                "No root path can be found for the provided "
                'module "%s".  This can happen because the '
                "module came from an import hook that does "
                "not provide file name information or because "
                "it's a namespace package.  In this case "
                "the root path needs to be explicitly "
                "provided." % import_name,
            )

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


class _PackageBoundObject(object):
    #: The name of the package or module that this app belongs to. Do not
    #: change this once it is set by the constructor.
    import_name = None

    #: Absolute path to the package on the filesystem. Used to look up
    #: resources contained in the package.
    root_path = None

    def __init__(self, import_name, template_folder=None, root_path=None):
        self.import_name = import_name

        if root_path is None:
            root_path = get_root_path(self.import_name)

        self.root_path = root_path



# Protean
import pytest

from protean.core.exceptions import IncorrectUsageError
from protean.core.repository import BaseRepository
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import Person, PersonRepository


class TestRepositoryInitialization:
    def test_that_base_repository_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            BaseRepository()

    def test_that_repository_can_be_instantiated(self, test_domain):
        repo = PersonRepository()
        assert repo is not None


class TestRepositoryRegistration:
    def test_that_repository_can_be_registered_with_domain(self, test_domain):
        test_domain.register(PersonRepository)

        assert (
            fully_qualified_name(PersonRepository) in test_domain.registry.repositories
        )

    def test_that_repository_can_be_registered_via_annotations(self, test_domain):
        @test_domain.repository
        class AnnotatedRepository:
            def special_method(self):
                pass

            class Meta:
                aggregate_cls = Person

        assert (
            fully_qualified_name(AnnotatedRepository)
            in test_domain.registry.repositories
        )

    def test_that_repository_can_be_registered_via_annotations_with_aggregate_cls_parameter(
        self, test_domain
    ):
        @test_domain.repository(aggregate_cls=Person)
        class AnnotatedRepository:
            def special_method(self):
                pass

        assert (
            fully_qualified_name(AnnotatedRepository)
            in test_domain.registry.repositories
        )

    def test_that_repository_cannot_be_registered_via_annotations_without_aggregate_cls(
        self, test_domain
    ):
        with pytest.raises(IncorrectUsageError):

            @test_domain.repository
            class AnnotatedRepository:
                def special_method(self):
                    pass

    def test_that_repository_can_be_retrieved_from_domain_by_its_aggregate_cls(
        self, test_domain
    ):
        test_domain.register(PersonRepository)

        assert isinstance(test_domain.repository_for(Person), PersonRepository)

    def test_that_fetching_an_unknown_repository_by_aggregate_cls_creates_one_automatically(
        self, test_domain
    ):
        repo = test_domain.repository_for(Person)
        assert repo.__class__.__name__ == "PersonRepository"



# Standard Library Imports
import importlib
import logging

# Protean
from protean.core.exceptions import ConfigurationError
from protean.globals import current_uow

logger = logging.getLogger("protean.email_providers")


class EmailProviders:
    def __init__(self, domain):
        self.domain = domain
        self._email_providers = None

    def _initialize_email_providers(self):
        """Read config file and initialize email providers"""
        configured_email_providers = self.domain.config["EMAIL_PROVIDERS"]
        email_provider_objects = {}

        if configured_email_providers and isinstance(configured_email_providers, dict):
            if "default" not in configured_email_providers:
                raise ConfigurationError("You must define a 'default' email provider")

            for provider_name, conn_info in configured_email_providers.items():
                provider_full_path = conn_info["PROVIDER"]
                provider_module, provider_class = provider_full_path.rsplit(
                    ".", maxsplit=1
                )

                provider_cls = getattr(
                    importlib.import_module(provider_module), provider_class
                )
                email_provider_objects[provider_name] = provider_cls(
                    provider_name, self, conn_info
                )

        self._email_providers = email_provider_objects

    def get_email_provider(self, provider_name):
        """Retrieve the email provider object with a given provider name"""
        if self._email_providers is None:
            self._initialize_email_providers()

        try:
            return self._email_providers[provider_name]
        except KeyError:
            raise AssertionError(f"No Provider registered with name {provider_name}")

    def send_email(self, email):
        """Push email through registered provider"""
        if self._email_providers is None:
            self._initialize_email_providers()

        if current_uow:
            logger.debug(
                f"Recording email {email.__class__.__name__} "
                f"to be sent to {repr(email)} in {current_uow}"
            )
            current_uow.register_email(email)
        else:
            logger.debug(
                f"Pushing {email.__class__.__name__} with content {repr(email)}"
            )
            self._email_providers[email.meta_.provider].send_email(email)



"""QuerySet Implementation"""

# Standard Library Imports
import copy
import logging

from typing import Any, Union

# Protean
from protean.utils.query import Q

logger = logging.getLogger("protean.repository.queryset")


class QuerySet:
    """A chainable class to gather a bunch of criteria and preferences (resultset size, order etc.)
    before execution.

    Internally, a QuerySet can be constructed, filtered, sliced, and generally passed around
    without actually fetching data. No data fetch actually occurs until you do something
    to evaluate the queryset.

    Once evaluated, a `QuerySet` typically caches its results. If the data in the database
    might have changed, you can get updated results for the same query by calling `all()` on a
    previously evaluated `QuerySet`.

    Attributes:
        offset: Number of records after which Results are fetched
        limit: The size the recordset to be pulled from database
        order_by: The list of parameters to be used for ordering the results.
            Use a `-` before the parameter name to sort in descending order
            and if not ascending order.
        excludes_: Objects with these properties will be excluded from the results
        filters: Filter criteria

    :return Returns a `ResultSet` object that holds the query results
    """

    def __init__(
        self,
        owner_dao,
        domain,
        entity_cls,
        criteria=None,
        offset: int = 0,
        limit: int = 10,
        order_by: set = None,
    ):
        """Initialize either with empty preferences (when invoked on an Entity)
            or carry forward filters and preferences when chained
        """
        self._owner_dao = owner_dao
        self._domain = domain
        self._entity_cls = entity_cls
        self._criteria = criteria or Q()
        self._result_cache = None
        self._offset = offset or 0
        self._limit = limit or 10

        # `order_by` could be empty, or a string or a list.
        #   Initialize empty list if `order_by` is None
        #   Convert string to list if `order_by` is a String
        #   Safe-cast list to a list if `order_by` is already a list
        if order_by:
            self._order_by = [order_by] if isinstance(order_by, str) else order_by
        else:
            self._order_by = []

        # `_temp_cache` is a data container for holding temporary
        #   It holds objects that have been added, but not yet persisted,
        #   as well as objects that have been removed, but are still
        #   present in the data store.
        #
        # The data set returned from Queryset is manipulated automatically
        #   to be up-to-date with temporary changes.
        self._temp_cache = {
            "added": list(),
            "removed": list(),
        }

    def _clone(self):
        """
        Return a copy of the current QuerySet.
        """
        clone = self.__class__(
            self._owner_dao,
            self._domain,
            self._entity_cls,
            criteria=self._criteria,
            offset=self._offset,
            limit=self._limit,
            order_by=self._order_by,
        )
        return clone

    #########################
    # Query support methods #
    #########################

    def _add_q(self, q_object):
        """Add a Q-object to the current filter."""
        self._criteria = self._criteria._combine(q_object, q_object.connector)

    def filter(self, *args, **kwargs):
        """
        Return a new QuerySet instance with the args ANDed to the existing
        set.
        """
        return self._filter_or_exclude(False, *args, **kwargs)

    def exclude(self, *args, **kwargs):
        """
        Return a new QuerySet instance with NOT (args) ANDed to the existing
        set.
        """
        return self._filter_or_exclude(True, *args, **kwargs)

    def _filter_or_exclude(self, negate, *args, **kwargs):
        clone = self._clone()
        if negate:
            clone._add_q(~Q(*args, **kwargs))
        else:
            clone._add_q(Q(*args, **kwargs))
        return clone

    def limit(self, limit):
        """Limit number of records"""
        clone = self._clone()

        if isinstance(limit, int):
            clone._limit = limit

        return clone

    def offset(self, offset):
        """Fetch results after `offset` value"""
        clone = self._clone()

        if isinstance(offset, int):
            clone._offset = offset

        return clone

    def order_by(self, order_by: Union[list, str]):
        """Update order_by setting for filter set"""
        clone = self._clone()
        if isinstance(order_by, str):
            order_by = [order_by]

        clone._order_by.extend(item for item in order_by if item not in clone._order_by)

        return clone

    def all(self):
        """Primary method to fetch data based on filters

        Also trigged when the QuerySet is evaluated by calling one of the following methods:
            * len()
            * bool()
            * list()
            * Iteration
            * Slicing
        """
        logger.debug(f"Query `{self.__class__.__name__}` objects with filters {self}")

        # Destroy any cached results
        self._result_cache = None

        # Fetch Model class and connected repository from Domain
        model_cls = self._domain.get_model(self._entity_cls)

        # order_by clause must be list of keys
        order_by = (
            self._entity_cls.meta_.order_by if not self._order_by else self._order_by
        )

        # Call the read method of the dao
        results = self._owner_dao._filter(
            self._criteria, self._offset, self._limit, order_by
        )

        # Convert the returned results to entity and return it
        entity_items = []
        for item in results.items:
            entity = model_cls.to_entity(item)
            entity.state_.mark_retrieved()
            entity_items.append(entity)
        results.items = entity_items

        # Cache results
        self._result_cache = results

        return results

    def update(self, *data, **kwargs):
        """Updates all objects with details given if they match a set of conditions supplied.

        This method updates each object individually, to fire callback methods and ensure
        validations are run.

        Returns the number of objects matched (which may not be equal to the number of objects
            updated if objects rows already have the new value).
        """
        updated_item_count = 0

        try:
            items = self.all()

            for item in items:
                self._owner_dao.update(item, *data, **kwargs)
                updated_item_count += 1
        except Exception:
            # FIXME Log Exception
            raise

        return updated_item_count

    def raw(self, query: Any, data: Any = None):
        """Runs raw query directly on the database and returns Entity objects

        Note that this method will raise an exception if the returned objects
            are not of the Entity type.

        `query` is not checked for correctness or validity, and any errors thrown by the plugin or
            database are passed as-is. Data passed will be transferred as-is to the plugin.

        All other query options like `order_by`, `offset` and `limit` are ignored for this action.
        """
        logger.debug(
            f"Query `{self.__class__.__name__}` objects with raw query {query}"
        )

        # Destroy any cached results
        self._result_cache = None

        # Fetch Model class and connected repository from Domain
        model_cls = self._domain.get_model(self._entity_cls)

        try:
            # Call the raw method of the repository
            results = self._owner_dao._raw(query, data)

            # Convert the returned results to entity and return it
            entity_items = []
            for item in results.items:
                entity = model_cls.to_entity(item)
                entity.state_.mark_retrieved()
                entity_items.append(entity)
            results.items = entity_items

            # Cache results
            self._result_cache = results
        except Exception:
            # FIXME Log Exception
            raise

        return results

    def delete(self):
        """Deletes matching objects from the Repository

        Does not throw error if no objects are matched.

        Returns the number of objects matched (which may not be equal to the number of objects
            deleted if objects rows already have the new value).
        """
        # Fetch Model class and connected repository from Domain
        deleted_item_count = 0

        try:
            items = self.all()

            for item in items:
                self._owner_dao.delete(item)
                deleted_item_count += 1
        except Exception:
            # FIXME Log Exception
            raise

        return deleted_item_count

    def update_all(self, *args, **kwargs):
        """Updates all objects with details given if they match a set of conditions supplied.

        This method forwards filters and updates directly to the repository. It does not
        instantiate entities and it does not trigger Entity callbacks or validations.

        Update values can be specified either as a dict, or keyword arguments.

        Returns the number of objects matched (which may not be equal to the number of objects
            updated if objects rows already have the new value).
        """
        updated_item_count = 0

        try:
            updated_item_count = self._owner_dao._update_all(
                self._criteria, *args, **kwargs
            )
        except Exception:
            # FIXME Log Exception
            raise

        return updated_item_count

    def delete_all(self, *args, **kwargs):
        """Deletes objects that match a set of conditions supplied.

        This method forwards filters directly to the repository. It does not instantiate entities and
        it does not trigger Entity callbacks or validations.

        Returns the number of objects matched and deleted.
        """
        deleted_item_count = 0
        try:
            deleted_item_count = self._owner_dao._delete_all(self._criteria)
        except Exception:
            # FIXME Log Exception
            raise

        return deleted_item_count

    ###############################
    # Python Magic method support #
    ###############################

    @property
    def _data(self):
        active_data = self._result_cache if self._result_cache else self.all()
        temp_data = copy.deepcopy(active_data)

        # Add objects in temporary cache
        for item in self._temp_cache["added"]:
            temp_data.items.append(item)
            temp_data.total += 1

        # Remove objects in temporary cache
        for item in self._temp_cache["removed"]:
            temp_data.items[:] = [
                value for value in temp_data.items if value.id != item.id
            ]
            temp_data.total -= 1

        return temp_data

    def __iter__(self):
        """Return results on iteration"""
        return iter(self._data)

    def __len__(self):
        """Return length of results"""
        return self._data.total

    def __bool__(self):
        """Return True if query results have items"""
        return bool(self._data)

    def __repr__(self):
        """Support friendly print of query criteria"""
        return "<%s: entity: %s, criteria: %s, offset: %s, limit: %s, order_by: %s>" % (
            self.__class__.__name__,
            self._entity_cls,
            self._criteria.deconstruct(),
            self._offset,
            self._limit,
            self._order_by,
        )

    def __getitem__(self, k):
        """Support slicing of results"""
        return self._data.items[k]

    def __contains__(self, k):
        """Support `in` operations"""
        return k.id in [item.id for item in self._data.items]

    #########################
    # Result properties #
    #########################

    @property
    def total(self):
        """Return the total number of records"""
        return self._data.total

    @property
    def items(self):
        """Return result values"""
        return self._data.items

    @property
    def first(self):
        """Return the first result"""
        return self._data.first

    @property
    def has_next(self):
        """Return True if there are more values present"""
        return self._data.has_next

    @property
    def has_prev(self):
        """Return True if there are previous values present"""
        return self._data.has_prev

    #######################
    # Association support #
    #######################

    def add(self, item):
        if item.id not in [value.id for value in self._data.items] or (
            item.id in [value.id for value in self._data.items]
            and item.state_.is_persisted
            and item.state_.is_changed
        ):
            if item.id not in [value.id for value in self._temp_cache["added"]]:
                # FIXME Re-evaluate for UoW support

                # If the child was already present, first remove that record
                if item.id in [value.id for value in self._data.items]:
                    for value in self._data.items:
                        if value.id == item.id:
                            self._temp_cache["removed"].append(value)
                            break

                # This updates the parent's unique identifier in the child
                #   so that the foreign key relationship is preserved
                for criteria in self._criteria.children:
                    setattr(item, criteria[0], criteria[1])
                self._temp_cache["added"].append(item)

    def remove(self, item):
        if item.id in [value.id for value in self._data.items]:
            if item.id not in [value.id for value in self._temp_cache["removed"]]:
                self._temp_cache["removed"].append(item)



# Standard Library Imports
from datetime import datetime
from typing import List

# Protean
from protean.core.aggregate import BaseAggregate
from protean.core.entity import BaseEntity
from protean.core.field.association import HasMany, HasOne, Reference
from protean.core.field.basic import Auto, DateTime, Integer, String, Text
from protean.core.repository import BaseRepository


class Role(BaseAggregate):
    name = String(max_length=15, required=True)
    created_on = DateTime(default=datetime.today())


class Person(BaseAggregate):
    first_name = String(max_length=50, required=True)
    last_name = String(max_length=50, required=True)
    age = Integer(default=21)


class PersonRepository(BaseRepository):
    def find_adults(self, age: int = 21) -> List[Person]:
        pass  # FIXME Implement filter method


# Aggregates to test Identity
class PersonAutoSSN(BaseAggregate):
    ssn = Auto(identifier=True)
    name = String(max_length=25)


class PersonExplicitID(BaseAggregate):
    ssn = String(max_length=36, identifier=True)
    name = String(max_length=25)


# Aggregates to test Subclassing
class SubclassRole(Role):
    pass


# Aggregates to test Abstraction # START #
class AbstractRole(BaseAggregate):
    foo = String(max_length=25)

    class Meta:
        abstract = True


class ConcreteRole(AbstractRole):
    bar = String(max_length=25)


class FurtherAbstractRole(ConcreteRole):
    foobar = String(max_length=25)

    class Meta:
        abstract = True


# Aggregates to test Abstraction # END #


# Aggregates to test Meta Info overriding # START #
class DbRole(BaseAggregate):
    bar = String(max_length=25)

    class Meta:
        schema_name = "foosball"


class SqlRole(Role):
    class Meta:
        schema_name = "roles"


class DifferentDbRole(Role):
    class Meta:
        provider = "non-default"


class SqlDifferentDbRole(Role):
    class Meta:
        provider = "non-default-sql"


class OrderedRole(BaseAggregate):
    bar = String(max_length=25)

    class Meta:
        order_by = "bar"


class OrderedRoleSubclass(Role):
    bar = String(max_length=25)

    class Meta:
        order_by = "bar"


# Aggregates to test Meta Info overriding # END #


# Aggregates to test associations # START #
class Post(BaseAggregate):
    content = Text(required=True)
    comments = HasMany("tests.aggregate.elements.Comment")
    author = Reference("tests.aggregate.elements.Author")


class PostVia(BaseAggregate):
    content = Text(required=True)
    comments = HasMany("tests.aggregate.elements.CommentVia", via="posting_id")
    author = Reference("tests.aggregate.elements.Author")


class PostViaWithReference(BaseAggregate):
    content = Text(required=True)
    comments = HasMany(
        "tests.aggregate.elements.CommentViaWithReference", via="posting_id"
    )
    author = Reference("tests.aggregate.elements.Author")


class Comment(BaseEntity):
    content = Text()
    added_on = DateTime()

    post = Reference("tests.aggregate.elements.Post")

    class Meta:
        aggregate_cls = Post


class CommentVia(BaseEntity):
    content = Text()
    added_on = DateTime()
    posting_id = String()

    class Meta:
        aggregate_cls = PostVia


class CommentViaWithReference(BaseEntity):
    content = Text()
    added_on = DateTime()
    posting = Reference("tests.aggregate.elements.PostVia")

    class Meta:
        aggregate_cls = PostViaWithReference


class Account(BaseAggregate):
    email = String(required=True, max_length=255, unique=True, identifier=True)
    password = String(required=True, max_length=255)
    username = String(max_length=255, unique=True)
    author = HasOne("tests.aggregate.elements.Author")


class Author(BaseEntity):
    first_name = String(required=True, max_length=25)
    last_name = String(max_length=25)
    posts = HasMany("tests.aggregate.elements.Post")
    account = Reference("tests.aggregate.elements.Account")

    class Meta:
        aggregate_cls = Account


class AccountWithId(BaseAggregate):
    email = String(required=True, max_length=255, unique=True)
    password = String(required=True, max_length=255)
    username = String(max_length=255, unique=True)
    author = HasOne("tests.aggregate.elements.Author")


class AccountVia(BaseAggregate):
    email = String(required=True, max_length=255, unique=True, identifier=True)
    password = String(required=True, max_length=255)
    username = String(max_length=255, unique=True)
    profile = HasOne("tests.aggregate.elements.ProfileVia", via="account_email")


class AccountViaWithReference(BaseAggregate):
    email = String(required=True, max_length=255, unique=True, identifier=True)
    password = String(required=True, max_length=255)
    username = String(max_length=255, unique=True)
    profile = HasOne("tests.aggregate.elements.ProfileViaWithReference", via="ac_email")


class Profile(BaseAggregate):
    about_me = Text()
    account = Reference("tests.aggregate.elements.Account", via="username")


class ProfileWithAccountId(BaseAggregate):
    about_me = Text()
    account = Reference("tests.aggregate.elements.AccountWithId")


class ProfileVia(BaseAggregate):
    profile_id = String(identifier=True)
    about_me = Text()
    account_email = String(max_length=255)


class ProfileViaWithReference(BaseAggregate):
    about_me = Text()
    ac = Reference("tests.aggregate.elements.AccountViaWithReference")


# Aggregates to test associations # END #



# Standard Library Imports
from datetime import datetime

# Protean
import pytest

from protean.core.aggregate import BaseAggregate
from protean.core.field.basic import DateTime, Dict, String
from sqlalchemy import types as sa_types


class Event(BaseAggregate):
    name = String(max_length=255)
    created_at = DateTime(default=datetime.utcnow())
    payload = Dict()


@pytest.mark.sqlite
def test_json_data_type_association(test_domain):
    test_domain.register(Event)

    model_cls = test_domain.get_model(Event)
    type(model_cls.payload.property.columns[0].type) == sa_types.PickleType


@pytest.mark.sqlite
def test_basic_array_data_type_operations(test_domain):
    test_domain.register(Event)

    model_cls = test_domain.get_model(Event)

    event = Event(
        name="UserCreated", payload={"email": "john.doe@gmail.com", "password": "*****"}
    )
    event_model_obj = model_cls.from_entity(event)

    event_copy = model_cls.to_entity(event_model_obj)
    assert event_copy is not None
    assert event_copy.payload == {"email": "john.doe@gmail.com", "password": "*****"}



"""Module to setup Factories and other required artifacts for tests

    isort:skip_file
"""
import os

import pytest


def pytest_addoption(parser):
    """Additional options for running tests with pytest"""
    parser.addoption(
        "--slow", action="store_true", default=False, help="Run slow tests"
    )
    parser.addoption(
        "--pending", action="store_true", default=False, help="Show pending tests"
    )
    parser.addoption(
        "--sqlite", action="store_true", default=False, help="Run Sqlite tests"
    )
    parser.addoption(
        "--postgresql", action="store_true", default=False, help="Run Postgresql tests"
    )
    parser.addoption(
        "--elasticsearch",
        action="store_true",
        default=False,
        help="Run Elasticsearch tests",
    )
    parser.addoption(
        "--redis", action="store_true", default=False, help="Run Redis based tests"
    )
    parser.addoption(
        "--sendgrid", action="store_true", default=False, help="Run Sendgrid tests"
    )


def pytest_collection_modifyitems(config, items):
    """Configure special markers on tests, so as to control execution"""
    run_slow = (
        run_pending
    ) = (
        run_sqlite
    ) = run_postgresql = run_elasticsearch = run_redis = run_sendgrid = False

    if config.getoption("--slow"):
        # --slow given in cli: do not skip slow tests
        run_slow = True

    if config.getoption("--pending"):
        run_pending = True

    if config.getoption("--sqlite"):
        run_sqlite = True

    if config.getoption("--postgresql"):
        run_postgresql = True

    if config.getoption("--elasticsearch"):
        run_elasticsearch = True

    if config.getoption("--redis"):
        run_redis = True

    if config.getoption("--sendgrid"):
        run_sendgrid = True

    skip_slow = pytest.mark.skip(reason="need --slow option to run")
    skip_pending = pytest.mark.skip(reason="need --pending option to run")
    skip_sqlite = pytest.mark.skip(reason="need --sqlite option to run")
    skip_postgresql = pytest.mark.skip(reason="need --postgresql option to run")
    skip_elasticsearch = pytest.mark.skip(reason="need --elasticsearch option to run")
    skip_redis = pytest.mark.skip(reason="need --redis option to run")
    skip_sendgrid = pytest.mark.skip(reason="need --sendgrid option to run")

    for item in items:
        if "slow" in item.keywords and run_slow is False:
            item.add_marker(skip_slow)
        if "pending" in item.keywords and run_pending is False:
            item.add_marker(skip_pending)
        if "sqlite" in item.keywords and run_sqlite is False:
            item.add_marker(skip_sqlite)
        if "postgresql" in item.keywords and run_postgresql is False:
            item.add_marker(skip_postgresql)
        if "elasticsearch" in item.keywords and run_elasticsearch is False:
            item.add_marker(skip_elasticsearch)
        if "redis" in item.keywords and run_redis is False:
            item.add_marker(skip_redis)
        if "sendgrid" in item.keywords and run_sendgrid is False:
            item.add_marker(skip_sendgrid)


@pytest.fixture(autouse=True)
def test_domain():
    from protean.domain import Domain

    domain = Domain("Test")

    # Construct relative path to config file
    current_path = os.path.abspath(os.path.dirname(__file__))
    config_path = os.path.join(current_path, "./config.py")

    if os.path.exists(config_path):
        domain.config.from_pyfile(config_path)

    with domain.domain_context():
        yield domain


@pytest.fixture(autouse=True)
def run_around_tests(test_domain):

    yield

    if test_domain.providers.has_provider("default"):
        test_domain.get_provider("default")._data_reset()



"""Module to test other Protean Functions and Utilities"""

# Protean
import pytest

from protean.core import entity
from protean.utils.importlib import perform_import


def test_perform_import():
    """ Test the perform import function """

    # Test importing of None
    mod = perform_import(None)
    assert mod is None

    # Test import of string
    mod = perform_import("protean.core.entity")
    assert mod == entity

    # Test import list
    mod = perform_import(
        ["protean.core.entity.BaseEntity", "protean.core.entity._EntityMetaclass"]
    )
    assert mod == [entity.BaseEntity, entity._EntityMetaclass]

    # Test Failed import
    with pytest.raises(ImportError):
        perform_import("protean.core.entity.xxxx")

    # Test Direct import
    mod = perform_import(entity)
    assert mod == entity



# Protean
import pytest

# Local/Relative Imports
from .elements import Person, PersonRepository, User


class TestDAO:
    """This class holds tests for DAO class"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)
        test_domain.register(PersonRepository, aggregate_cls=Person)
        test_domain.register(User)

    def test_successful_initialization_of_dao(self, test_domain):
        test_domain.get_dao(Person).query.all()
        provider = test_domain.get_provider("default")
        conn = provider.get_connection()
        assert isinstance(conn._db["data"], dict)

    @pytest.mark.xfail
    def test_that_fields_can_have_custom_attribute_names(self, test_domain):
        dao = test_domain.get_dao(Person)
        person1 = dao.create(id=1, first_name="Athos", last_name="Musketeer", age=2)

        model = dao.model_cls.from_entity(person1)
        assert all(attribute in model for attribute in ["prenom", "nom_de_famille"])

        entity = dao.model_cls.to_entity(model)
        assert all(
            field_name in entity.to_dict() for field_name in ["first_name", "last_name"]
        )

    def test_that_escaped_quotes_in_values_are_handled_properly(self, test_domain):
        test_domain.get_dao(Person).create(
            id=1, first_name="Athos", last_name="Musketeer", age=2
        )
        test_domain.get_dao(Person).create(
            id=2, first_name="Porthos", last_name="Musketeer", age=3
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Aramis", last_name="Musketeer", age=4
        )

        person1 = test_domain.get_dao(Person).create(
            first_name="d'Artagnan1", last_name="John", age=5
        )
        person2 = test_domain.get_dao(Person).create(
            first_name="d'Artagnan2", last_name="John", age=5
        )
        person3 = test_domain.get_dao(Person).create(
            first_name='d"Artagnan3', last_name="John", age=5
        )
        person4 = test_domain.get_dao(Person).create(
            first_name='d"Artagnan4', last_name="John", age=5
        )

        assert all(
            person is not None for person in [person1, person2, person3, person4]
        )



# Standard Library Imports
import logging
import logging.config

from collections.abc import Iterable

# Protean
from celery import Celery, Task
from kombu import Queue
from protean.core.domain_event import BaseDomainEvent
from protean.port.broker import BaseBroker
from protean.utils import DomainObjects, fully_qualified_name
from protean.utils.inflection import underscore

logger = logging.getLogger("protean.adapters.celery")


class ProteanTask(Task):
    """The default base class for all Task classes constructed from Subscribers/Command Handlers.
    """

    pass


class CeleryBroker(BaseBroker):
    def __init__(self, name, domain, conn_info):
        super().__init__(name, domain, conn_info)
        self.celery_app = Celery(broker=conn_info["URI"], backend=conn_info["URI"],)

        self.celery_app.conf.update(enable_utc=True)

        # We construct queues dynamically when subscribers register
        self.queues = []

    def construct_and_register_celery_task(self, consumer_cls):
        """Constructs a Celery-compliant Task class and also registers
        Task with Celery App

        Arguments:
            consumer_cls {BaseSubscriber} -- The Subscriber or Command Handler class
                                             to be converted into a Celery Task

        Returns:
            ProteanTask -- Decorated and Registered Celery Task class
        """
        attrs = consumer_cls.__dict__
        custom_attrs = {
            "run": attrs["notify"],  # `notify` is the method to run on event
            "name": underscore(
                fully_qualified_name(consumer_cls)
            ),  # `name` will be the same as the task's queue
        }
        attrs = {**attrs, **custom_attrs}

        # Construct `decorated_cls` dynamically from `ProteanTask`.
        #   `ProteanTask` acts as the base class for all celery tasks.
        decorated_cls = type(consumer_cls.__name__ + "Task", (ProteanTask,), {**attrs})

        # Register Task class with Celery app
        decorated_cls_instance = self.celery_app.register_task(decorated_cls())

        # Add to Queue so that workers pick it up automatically
        self.queues.append(Queue(decorated_cls.name))

        return decorated_cls_instance

    def register(self, initiator_cls, consumer_cls):
        """Registers Domain Events and Commands with Subscribers/Command Handlers

        Arguments:
            initiator_cls {list} -- One or more Domain Events or Commands
            consumer_cls {Subscriber/CommandHandler} -- The consumer class connected to the Domain Event or Command
        """
        if not isinstance(initiator_cls, Iterable):
            initiator_cls = [initiator_cls]

        decorated_cls_instance = self.construct_and_register_celery_task(consumer_cls)

        for initiator in initiator_cls:
            if initiator.element_type == DomainObjects.DOMAIN_EVENT:
                self._subscribers[fully_qualified_name(initiator)].add(
                    decorated_cls_instance
                )
                logger.debug(
                    f"Registered Subscriber {decorated_cls_instance.__class__.__name__} with queue "
                    "{self.celery_app.tasks} as Celery Task"
                )
            else:
                self._command_handlers[
                    fully_qualified_name(initiator)
                ] = decorated_cls_instance

    def send_message(self, initiator_obj):
        if isinstance(initiator_obj, BaseDomainEvent):
            for subscriber in self._subscribers[
                fully_qualified_name(initiator_obj.__class__)
            ]:
                if self.conn_info["IS_ASYNC"]:
                    subscriber.apply_async(
                        [initiator_obj.to_dict()], queue=subscriber.name
                    )
                else:
                    subscriber.apply([initiator_obj.to_dict()])
        else:
            raise NotImplementedError



# cSpell: disable

# Protean
from protean.utils import Database, IdentityStrategy, IdentityType

####################
# CORE             #
####################

DEBUG = True
TESTING = True
ENV = "development"

# A secret key for this particular Protean installation. Used in secret-key
# hashing algorithms.
SECRET_KEY = "nU5JSWCP#4c#Annek2mx9V&g5uWUJfh@"

IDENTITY_STRATEGY = IdentityStrategy.UUID
IDENTITY_TYPE = IdentityType.UUID

# Define the databases
DATABASES = {
    "default": {
        "PROVIDER": "protean.adapters.ESProvider",
        "DATABASE": Database.ELASTICSEARCH.value,
        "DATABASE_URI": {"hosts": ["localhost"]},
    },
}



""" Test the Q object used for managing filter criteria """
# Protean
import pytest

from protean.core.queryset import QuerySet
from protean.utils.query import Q

# Local/Relative Imports
from .elements import Person


class TestCriteriaConstruction:
    """Test Conjunction operations on QuerySet"""

    @pytest.fixture(autouse=True)
    def register_elements(self, test_domain):
        test_domain.register(Person)

    def test_that_an_empty_Q_object_is_initialized_with_queryset(self, test_domain):
        assert isinstance(test_domain.get_dao(Person).query, QuerySet)
        assert isinstance(test_domain.get_dao(Person).query._criteria, Q)

    def test_query_construction_with_simple_filter_kwargs(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(last_name="John")
        assert q1._criteria is not None

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert decon_args == ()
        assert decon_kwargs == {"last_name": "John"}
        assert q1._criteria.connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_simple_exclude_kwargs(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.exclude(last_name="John")

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert decon_args == ()
        assert decon_kwargs == {"last_name": "John", "_negated": True}
        assert q1._criteria.connector == Q.AND
        assert q1._criteria.negated is True
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_multiple_filter_criteria(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(last_name="John")
        q2 = q1.filter(age=3)

        _, decon_args, decon_kwargs = q2._criteria.deconstruct()

        assert decon_args == (("last_name", "John"), ("age", 3))
        assert decon_kwargs == {}
        assert q1._criteria.connector == Q.AND
        assert q2._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q2._criteria

    def test_query_construction_with_multiple_exclude_criteria(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.exclude(last_name="John")
        q2 = q1.exclude(age=3)

        _, decon_args, decon_kwargs = q2._criteria.deconstruct()

        assert str(decon_args) == (
            "(<Q: (NOT (AND: ('last_name', 'John')))>, " "<Q: (NOT (AND: ('age', 3)))>)"
        )
        assert decon_kwargs == {}
        assert q1._criteria.connector == Q.AND
        assert q2._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q2._criteria

    def test_query_construction_with_multiple_criteria_in_filter(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(last_name="John", age=3)
        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert decon_args == (("age", 3), ("last_name", "John"))
        assert decon_kwargs == {}
        assert q1._criteria.connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_multiple_criteria_in_exclude(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.exclude(last_name="John", age=3)

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert decon_args == (("age", 3), ("last_name", "John"))
        assert decon_kwargs == {"_negated": True}
        assert q1._criteria.connector == Q.AND
        assert q1._criteria.negated is True
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_combined_filter_and_exclude_with_filter_coming_first(
        self, test_domain
    ):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(last_name="John").exclude(age=3)

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert (
            str(decon_args) == "(('last_name', 'John'), <Q: (NOT (AND: ('age', 3)))>)"
        )
        assert decon_kwargs == {}
        assert q1._criteria.connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_combined_filter_and_exclude_with_exclude_coming_first(
        self, test_domain
    ):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.exclude(age=3).filter(last_name="John")

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert (
            str(decon_args) == "(<Q: (NOT (AND: ('age', 3)))>, ('last_name', 'John'))"
        )
        assert decon_kwargs == {}
        assert q1._criteria.children[0].connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_single_Q_object_input_to_filter_method(
        self, test_domain
    ):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(Q(last_name="John"))

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert str(decon_args) == "(<Q: (AND: ('last_name', 'John'))>,)"
        assert decon_kwargs == {}
        assert q1._criteria.children[0].connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_single_Q_object_to_exclude_method(
        self, test_domain
    ):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.exclude(Q(last_name="John"))

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert str(decon_args) == "(<Q: (AND: ('last_name', 'John'))>,)"
        assert decon_kwargs == {"_negated": True}
        assert q1._criteria.children[0].connector == Q.AND
        assert q1._criteria.negated is True
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construnction_with_multiple_Q_objects_input_to_filter_method(
        self, test_domain
    ):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(Q(last_name="John"), Q(age=3))

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert (
            str(decon_args)
            == "(<Q: (AND: ('last_name', 'John'))>, <Q: (AND: ('age', 3))>)"
        )
        assert decon_kwargs == {}
        assert q1._criteria.children[0].connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construnction_with_multiple_Q_objects_input_to_exclude_method(
        self, test_domain
    ):
        q1 = test_domain.get_dao(Person).query.exclude(Q(last_name="John"), Q(age=3))

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert (
            str(decon_args)
            == "(<Q: (AND: ('last_name', 'John'))>, <Q: (AND: ('age', 3))>)"
        )
        assert decon_kwargs == {"_negated": True}
        assert q1._criteria.children[0].connector == Q.AND
        assert q1._criteria.negated is True
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_AND_filter(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(Q(last_name="John") & Q(age=3))

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert str(decon_args) == "(<Q: (AND: ('last_name', 'John'), ('age', 3))>,)"
        assert decon_kwargs == {}
        assert q1._criteria.children[0].connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_OR_filter(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(Q(last_name="John") | Q(age=3))

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert str(decon_args) == "(<Q: (OR: ('last_name', 'John'), ('age', 3))>,)"
        assert decon_kwargs == {}
        assert q1._criteria.children[0].connector == Q.OR
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_multiple_AND_criteria(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(
            Q(last_name="John") & Q(age=3) & Q(name="Jean")
        )

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert str(decon_args) == (
            "(<Q: (AND: "
            "('last_name', 'John'), "
            "('age', 3), "
            "('name', 'Jean'))>,)"
        )
        assert decon_kwargs == {}
        assert q1._criteria.children[0].connector == Q.AND
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_multiple_OR_criteria(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(
            Q(last_name="John") | Q(age=3) | Q(name="Jean")
        )

        _, decon_args, decon_kwargs = q1._criteria.deconstruct()

        assert str(decon_args) == (
            "(<Q: (OR: " "('last_name', 'John'), " "('age', 3), " "('name', 'Jean'))>,)"
        )
        assert decon_kwargs == {}
        assert q1._criteria.children[0].connector == Q.OR
        assert q1._criteria.negated is False
        assert Q(*decon_args, **decon_kwargs) == q1._criteria

    def test_query_construction_with_AND_OR_combinations(self, test_domain):
        # Filter by the last_name
        q1 = test_domain.get_dao(Person).query.filter(
            Q(last_name="John") | Q(age=3), first_name="Jean"
        )
        _, decon_args, decon_kwargs = q1._criteria.deconstruct()
        assert str(decon_args) == (
            "(<Q: (OR: ('last_name', 'John'), ('age', 3))>, " "('first_name', 'Jean'))"
        )
        assert Q(*decon_args, **decon_kwargs) == q1._criteria
        assert q1._criteria.children[0].connector == Q.OR

        q2 = test_domain.get_dao(Person).query.filter(
            (Q(last_name="John") | Q(age=3)), Q(first_name="Jean")
        )
        _, decon_args, decon_kwargs = q2._criteria.deconstruct()
        assert str(decon_args) == (
            "(<Q: (OR: ('last_name', 'John'), ('age', 3))>, "
            "<Q: (AND: ('first_name', 'Jean'))>)"
        )
        assert Q(*decon_args, **decon_kwargs) == q2._criteria
        assert q2._criteria.children[0].connector == Q.OR

        q3 = test_domain.get_dao(Person).query.filter(
            Q(first_name="Jean") & (Q(last_name="John") | Q(age=3))
        )
        _, decon_args, decon_kwargs = q3._criteria.deconstruct()
        assert str(decon_args) == (
            "(<Q: (AND: "
            "('first_name', 'Jean'), "
            "(OR: ('last_name', 'John'), ('age', 3)))>,)"
        )
        assert Q(*decon_args, **decon_kwargs) == q3._criteria
        assert q3._criteria.children[0].connector == Q.AND

    def test_clone(self, test_domain):
        """Test that clone works as expected... it clones!"""
        query1 = test_domain.get_dao(Person).query.filter(last_name="John")
        query2 = query1.filter(age=3)
        query3 = query2.order_by("name")

        assert query1 != query2
        assert query2 != query3

    def test_list(self, test_domain):
        """Test that filter is evaluted on calling `list()`"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        dogs = list(query)

        assert dogs is not None
        assert len(dogs) == 2

    def test_repr(self, test_domain):
        """Test that filter is evaluted on calling `list()`"""
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        assert repr(query) == (
            "<QuerySet: entity: <class 'tests.query.elements.Person'>, "
            "criteria: ('protean.utils.query.Q', (), {'last_name': 'John'}), "
            "offset: 0, "
            "limit: 10, order_by: ['age']>"
        )

    def test_bool_false(self, test_domain):
        """Test that `bool` returns `False` on no records"""
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        assert bool(query) is False

    def test_bool_true(self, test_domain):
        """Test that filter is evaluted on calling `list()`"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )

        assert bool(query) is True

    def test_len(self, test_domain):
        """Test that filter is evaluted on calling `list()`"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        assert len(query) == 2

    def test_slice(self, test_domain):
        """Test slicing on filter"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )
        test_domain.get_dao(Person).create(
            id=5, first_name="Fred", age=4, last_name="Constantine"
        )
        test_domain.get_dao(Person).create(
            id=6, first_name="Flint", age=2, last_name="Steve"
        )

        # Filter by Dog attributes
        query = test_domain.get_dao(Person).query.order_by("age")
        sliced = query[1:]
        assert len(sliced) == 4

    def test_caching(self, test_domain):
        """Test that results are cached after query is evaluated"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )

        # Result cache is empty to begin with
        assert query._result_cache is None

        # Total invokes an evaluation and a query
        assert query.total == 2

        # Result cache is now populated
        assert query._result_cache is not None
        assert query._result_cache.total == 2

    def test_cache_reset(self, test_domain):
        """Test that results are cached after query is evaluated"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )

        # Total invokes an evaluation and a query
        assert query.total == 2
        assert query._result_cache.total == 2

        query_dup = query.limit(25)
        assert query_dup._result_cache is None

    def test_total(self, test_domain):
        """Test value of `total` results"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        assert query.total == 2

    def test_total_with_cache(self, test_domain):
        """Test value of `total` results without refresh"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        assert query.total == 2

        test_domain.get_dao(Person).create(
            id=5, first_name="Berry", age=1, last_name="John"
        )
        assert query.total == 2
        assert query._result_cache.total == 2

        # Force a refresh
        assert query.all().total == 3

        # Result cache is now populated
        assert query._result_cache.total == 3

    def test_items(self, test_domain):
        """Test that items is retrieved from ResultSet"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        assert query.items[0].id == query.all().items[0].id

    def test_items_with_cache(self, test_domain):
        """Test that items is retrieved from ResultSet"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = (
            test_domain.get_dao(Person).query.filter(last_name="John").order_by("age")
        )
        assert query.items[0].id == 3

        test_domain.get_dao(Person).create(
            id=5, first_name="Berry", age=1, last_name="John"
        )
        assert query.items[0].id == 3

        assert query.all().items[0].id == 5

    def test_has_next(self, test_domain):
        """Test if there are results after the current set"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = test_domain.get_dao(Person).query.limit(2)
        assert query.has_next is True

    def test_has_next_with_cache(self, test_domain):
        """Test if there are results after the current set"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        dog = test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = test_domain.get_dao(Person).query.limit(2)
        assert query.has_next is True

        test_domain.get_dao(Person).delete(dog)

        assert query.has_next is True
        assert query.all().has_next is False

    def test_has_prev(self, test_domain):
        """Test if there are results before the current set"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = test_domain.get_dao(Person).query.offset(2).limit(2)
        assert query.has_prev is True

    def test_has_prev_with_cache(self, test_domain):
        """Test if there are results before the current set"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        dog = test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = test_domain.get_dao(Person).query.offset(2).limit(2)
        assert query.has_prev is True

        test_domain.get_dao(Person).delete(dog)

        assert query.has_prev is True
        assert query.all().has_prev is False

    def test_first(self, test_domain):
        """Test that the first item is retrieved correctly from the resultset"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = test_domain.get_dao(Person).query.order_by("-age")
        assert query.first.id == 2

    def test_first_with_cache(self, test_domain):
        """Test that the first item is retrieved correctly from the resultset"""
        # Add multiple entries to the DB
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        query = test_domain.get_dao(Person).query.order_by("-age")
        assert query.first.id == 2

        test_domain.get_dao(Person).create(
            id=5, first_name="Berry", age=8, last_name="John"
        )
        assert query.first.id == 2
        assert query.all().first.id == 5

    def test_raw(self, test_domain):
        """Test raw queries"""
        test_domain.get_dao(Person).create(
            id=2, first_name="Murdock", age=7, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=3, first_name="Jean", age=3, last_name="John"
        )
        test_domain.get_dao(Person).create(
            id=4, first_name="Bart", age=6, last_name="Carrie"
        )

        # Filter by Dog attributes
        results = test_domain.get_dao(Person).query.raw('{"last_name":"John"}')
        assert results.total == 2

        results = test_domain.get_dao(Person).query.raw("{'last_name':'John'}")
        assert results.total == 2

        results = test_domain.get_dao(Person).query.raw('{"last_name":"John", "age":3}')
        assert results.total == 1

        results = test_domain.get_dao(Person).query.raw(
            '{"last_name":"John", "age__in":[6,7]}'
        )
        assert results.total == 1

        results = test_domain.get_dao(Person).query.raw(
            '{"last_name":"John", "age__in":[3,7]}'
        )
        assert results.total == 2



# Standard Library Imports
from functools import partial

# Protean
from werkzeug.local import LocalProxy, LocalStack

_domain_ctx_err_msg = """\
Working outside of domain context.
This typically means that you attempted to use functionality that needed
to interface with the current domain object in some way. To solve
this, set up an domain context with domain.domain_context().  See the
documentation for more information.\
"""


def _lookup_domain_object(name):
    top = _domain_context_stack.top
    if top is None:
        raise RuntimeError(_domain_ctx_err_msg)
    return getattr(top, name)


def _find_domain():
    top = _domain_context_stack.top
    if top is None:
        raise RuntimeError(_domain_ctx_err_msg)
    return top.domain


def _find_uow():
    return _uow_context_stack.top


# context locals
_domain_context_stack = LocalStack()
_uow_context_stack = LocalStack()
current_domain = LocalProxy(_find_domain)
current_uow = LocalProxy(_find_uow)
g = LocalProxy(partial(_lookup_domain_object, "g"))



# Standard Library Imports
from collections import OrderedDict
from datetime import datetime

# Protean
import pytest

from protean.core.entity import _EntityMetaclass
from protean.core.exceptions import ValidationError
from protean.utils import fully_qualified_name

# Local/Relative Imports
from .elements import (
    AccountWithId,
    Comment,
    ConcreteRole,
    Person,
    Post,
    ProfileWithAccountId,
    Role,
    SubclassRole,
)


class TestAggregateStructure:
    def test_aggregate_inheritance(self):
        assert isinstance(Role, _EntityMetaclass)

    def test_successful_aggregate_registration(self, test_domain):
        test_domain.register(Role)
        assert fully_qualified_name(Role) in test_domain.registry.aggregates

    def test_field_definitions_declared_in_aggregate(self):
        declared_fields_keys = list(
            OrderedDict(sorted(Role.meta_.declared_fields.items())).keys()
        )
        assert declared_fields_keys == ["created_on", "id", "name"]

    def test_declared_reference_fields_in_an_aggregate(self, test_domain):
        declared_fields_keys = list(
            OrderedDict(sorted(Post.meta_.declared_fields.items())).keys()
        )
        assert declared_fields_keys == ["author", "comments", "content", "id"]

    def test_declared_has_one_fields_in_an_aggregate(self, test_domain):
        # `author` is a HasOne field, so it should be:
        #   absent in attributes
        #   present as `author` in declared_fields
        assert all(
            key in AccountWithId.meta_.declared_fields
            for key in ["author", "email", "id", "password"]
        )
        assert all(
            key in AccountWithId.meta_.attributes for key in ["email", "id", "password"]
        )
        assert "author" not in AccountWithId.meta_.attributes

        # `account` is a Reference field, so it should be present as:
        #   `account_id` in attributes
        #   `account` in declared_fields
        assert all(
            key in ProfileWithAccountId.meta_.declared_fields
            for key in ["about_me", "account"]
        )
        assert all(
            key in ProfileWithAccountId.meta_.attributes
            for key in ["about_me", "account_id"]
        )

    def test_declared_has_many_fields_in_an_aggregate(self, test_domain):
        # `comments` is a HasMany field, so it should be:
        #   absent in attributes
        #   present as `comments` in declared_fields
        assert all(
            key in Post.meta_.declared_fields
            for key in ["comments", "content", "id", "author"]
        )
        assert all(
            key in Post.meta_.attributes for key in ["content", "id", "author_id"]
        )
        assert "comments" not in Post.meta_.attributes

        # `post` is a Reference field, so it should be present as:
        #   `post_id` in attributes
        #   `post` in declared_fields
        assert all(
            key in Comment.meta_.declared_fields
            for key in ["added_on", "content", "id", "post"]
        )
        assert all(
            key in Comment.meta_.attributes
            for key in ["added_on", "content", "id", "post_id"]
        )


class TestSubclassedAggregateStructure:
    def test_subclass_aggregate_field_definitions(self):
        declared_fields_keys = list(
            OrderedDict(sorted(SubclassRole.meta_.declared_fields.items())).keys()
        )
        assert declared_fields_keys == ["created_on", "id", "name"]

    def test_that_fields_in_base_classes_are_inherited(self):
        declared_fields_keys = list(
            OrderedDict(sorted(ConcreteRole.meta_.declared_fields.items())).keys()
        )
        assert declared_fields_keys == ["bar", "foo", "id"]

        role = ConcreteRole(id=3, foo="foo", bar="bar")
        assert role is not None
        assert role.foo == "foo"


class TestAggregateInitialization:
    def test_successful_aggregate_initialization(self):
        role = Role(name="ADMIN")
        assert role is not None
        assert role.name == "ADMIN"
        assert type(role.created_on) is datetime

    def test_individuality(self):
        """Test successful Account Entity initialization"""

        role1 = Role(name="ADMIN")
        role2 = Role(name="USER")
        assert role1.name == "ADMIN"
        assert role2.name == "USER"

    def test_initialization_from_dict_template(self):
        with pytest.raises(AssertionError):
            Person("John Doe")

        person = Person({"first_name": "John", "last_name": "Doe", "age": 23})
        assert person.first_name == "John"
        assert person.last_name == "Doe"
        assert person.age == 23

    def test_error_message_content_on_validation_error(self):
        # Single error message
        try:
            Person(last_name="Doe")
        except ValidationError as err:
            assert err.messages == {"first_name": ["is required"]}

        # Test multiple error messages
        try:
            Person(last_name="Doe", age="old")
        except ValidationError as err:
            assert err.messages == {
                "first_name": ["is required"],
                "age": ['"old" value must be an integer.'],
            }


class TestAggregateFieldValues:
    def test_that_validation_error_is_raised_if_required_fields_are_not_provided(self):
        with pytest.raises(ValidationError):
            Role(id=123423)

    def test_that_field_values_are_defaulted_when_not_provided(self):
        """Test that values are defaulted properly"""
        person = Person(first_name="John", last_name="Doe")
        assert person.age == 21

    def test_that_field_values_with_default_settings_can_be_specified_explicitly(self):
        """Test that values are defaulted properly"""
        person = Person(first_name="John", last_name="Doe", age=35)
        assert person.age == 35

    def test_that_validation_error_is_raised_when_specified_string_length_is_breached(
        self,
    ):
        """Test validation of String length checks"""
        with pytest.raises(ValidationError):
            Role(name="THIS_IS_A_VERY_LONG_ROLE_NAME")

    def test_that_values_are_validated_against_specified_data_types(self):
        """Test validation of data types of values"""
        with pytest.raises(ValidationError):
            Person(first_name="John", age="Young")
